type uident = int

type use = 
    UOmega
  | UOne
  | UPlus
  | UVar of uident

type typ =
    TString
  | TInt
  | TBool
(*  | TPcdata *)
  | TLabel
  | TAttribute
  | TFun of typ * typ
  | TTree of use
  | TTreelist of use
  | TTuple of typ list
  | TTupleGet of typ * int


type ident = {id:int; origin:Syntax.ident}

let new_int =
  let counter = ref 1 in
    fun () -> incr counter; !counter

let new_id orig_id =
  {id = new_int (); origin=orig_id}

let new_uvar =
  let prev = ref (1) in
    (fun () -> incr prev; UVar !prev)

type label = string
type pcdata = string

type binop = Eq | Cat | Sub | ListCat | Semi | Mod

type pre_t =
    String of string
  | Int of int
  | True | False
  (* | Pcdata of pcdata *)
  | Attribute of attribute
  | Label of label
  | BinOp of binop * t * t
  | Var of ident
  | Fix of ident * ident * t
  | App of t * t
  | Leaf of t
  | Node of t * t * t
  | Nil
  | Cons of t * t
  | Match of typ * ident * one_match list
  | If of t * t * t
  | Let of ident * t * t
  | Letrecs of (ident * t) list * t
  | AttrGet of t * ident
and t = pre_t * typ
and attribute = (string * Syntax.t) list
and one_match = (pattern * t)
and pattern =
    LeafPattern of (*string*) pattern
  | NodePattern of (*label*) pattern * (*attribute*)pattern * (*treelist*)pattern
  | NilPattern
  | ConsPattern of (*tree*)pattern * (*treelist*)pattern
  | IdentPattern of ident
  | StringPattern of string
  | LabelPattern of label

let rec check_match p pat =
  match p, pat with
(*     (IdentPattern x, IdentPattern y) when x = y -> true
  | (IdentPattern {origin = "_"; id = _}, _) | (_, IdentPattern {origin = "_"; id = _}) -> true *)
  | IdentPattern _, IdentPattern _ -> true
  | StringPattern s1, StringPattern s2 when s1 = s2 -> true
  | LabelPattern l1, LabelPattern l2 when l1 = l2 -> true
  | ConsPattern (p_hd,p_tl), ConsPattern(pat_hd, pat_tl) ->
      (check_match p_hd pat_hd) && (check_match p_tl pat_tl)
  | NilPattern, NilPattern -> true
  | NodePattern(p1, p2, p3), NodePattern(p1', p2', p3') ->
      (check_match p1 p1') && (check_match p2 p2') && (check_match p3 p3')
  | LeafPattern p, LeafPattern p' -> check_match p p'
  | _ -> false

let rec pick_case p pats =
  match pats with
    [] -> failwith "No case matches with the given pattern"
  | (pat, t)::tl ->
      if check_match p pat then
	(pat, t), tl
      else
	let t_match, t_else = pick_case p tl in
	t_match, ((pat,t)::t_else)

(*
  Returns the list of variables that occur in the passed pattern.
*)
let rec pick_vars = function
    (NilPattern | LeafPattern _ | LabelPattern _ | StringPattern _) -> []
  | ConsPattern(p1, p2) -> (pick_vars p1) @ (pick_vars p2)
  | IdentPattern x -> [x]
  | NodePattern(p1, p2, p3) -> (* (pick_vars p1) @ (pick_vars p2) @ *) (pick_vars p1) @ (pick_vars p3)

let rec print_usage u =
  match u with
    UOmega -> print_string "Omega"
  | UOne -> print_string "One"
  | UPlus -> print_string "Plus"
  | UVar i -> print_string "u";print_int i
let rec print_typ typ = 
  match typ with
      TTree u -> print_string "TTree^{"; print_usage u; print_string "}" 
    | TTreelist u -> print_string "TTreelist^{"; print_usage u; print_string "}" 
    | TLabel -> print_string "TLabel"
    | TAttribute-> print_string "TAttr"
    | TFun(t1, t2)->
	print_string "TFun(";
	print_typ t1;
	print_string ", ";
	print_typ t2;
	print_string ")"	
    | TString -> print_string "TStr"
    | TInt -> print_string "TInt"
    | TLabel -> print_string "TLabel"
    | _ -> print_string "other_typ"
let rec print_id id =
  (* print_int id.id *)
  print_string id.origin
let rec print_ids ids =
  match ids with
      [] -> print_string "[]"
    | id::ids' ->
	begin
	  let rec inner ids =
	    match ids with
		hd::tl ->
		  print_string ", ";
		  print_id hd;
		  inner tl
	      | [] -> ()
	  in
	    print_string "[";
	    print_id id;
	    inner ids';
	    print_string "]";
	end
let print_binop = function 
      Eq -> print_string "="
    | Sub -> print_string "-"
    | Cat -> print_string "^"
    | ListCat -> print_string "@@"
    | Semi -> print_string ";"
    | Mod -> print_string "/"

let rec print_t t =
  match t with
      String s, _ -> print_string s
    | Int n, _ -> print_int n
    (* | Pcdata pc, _  -> print_string "pc" *)
    | Label l, _  -> print_string "label"
    | True, _ -> print_string "true"
    | False, _ -> print_string "false"
    | Attribute a, _ -> print_string "attr"
    | Var id, _ -> print_id id
    | Fix (f, x, t1), _ ->
	print_string "fix(";
	print_id f;
	print_string ", ";
	print_id x;
	print_string ", ";
	print_t t1;
	print_string ")"
    | BinOp(binop, t1, t2), _  ->
	print_string "(";
	print_t t1;
	print_binop binop;
	print_t t2;
	print_string ")";
    | Let(x, t1, t2), _  ->
	print_string "Let(";
	print_id x;
	print_string ", ";
	print_t t1;
	print_string ", ";
	print_t t2;
	print_string ")"
    | Letrecs(fts, t), _ ->
	print_string "letrecs";
	let rec tmp fts =
	   (match fts with
	       (f, t)::tl ->
		 print_string "(";
		 print_id f;
		 print_string ", ";
		 print_t t;
		 print_string ")";
		 tmp tl
	      | [] -> ())
	in
	  print_string "(";
	  tmp fts;
	  print_string ", ";
	  print_t t;
	  print_string ")"	   
    | App (t1, t2), _  -> print_string "app(";
	print_t t1; print_string ", "; print_t t2; print_string ")"
    | If (t0, t1, t2), _ ->
	print_string "if ";
	print_t t0;
	print_string " then ";
	print_t t1;
	print_string " else ";
	print_t t2
    | Leaf t, _ -> print_string "leaf "; print_t t
    | Node(t1, t2, t3), _ ->
	print_string "node(";
	print_t t1; print_string ", ";
	print_t t2; print_string ", ";
	print_t t3; print_string ")"
    | Cons(t1, t2), _ ->
	print_string "cons(";
	print_t t1; print_string ", ";
	print_t t2; print_string ")"
    | Nil, _ -> print_string "nil"
    | Match (typ, id, matchs), _ ->
	print_string "match ";
	print_typ typ;
	print_id id;
	print_string " with \n";
	print_matchs matchs
    | _ -> print_string "other"
and print_matchs ms =
  let rec print_pattern p =
    match p with
	LeafPattern s ->
	  print_string "PC ";
	  print_pattern s
      | NodePattern (p1, p2, p3) ->
	  print_pattern p1;
	  print_string "@";
	  print_pattern p2;
	  print_string "$";
	  print_pattern p3
      | NilPattern ->
	  print_string "[]"
      | ConsPattern (p1,p2)-> 
	  print_pattern p1;
	  print_string "::";
	  print_pattern p2
      | StringPattern s ->
	  print_string s
      | LabelPattern l ->
	  print_string ("<"^l^">")
      | IdentPattern id -> 
	  print_id id
  in
  let rec print_one_match (p, t) =
    print_pattern p;
    print_string " -> ";
    print_t t;
    print_newline()
  in
  match ms with
      [] -> ()
    | hd::tl ->
	print_one_match hd;
	print_matchs tl







let rec subtract id ids =
  List.filter (fun x -> x <> id) ids

let rec subtracts ids1 ids2 =
  match ids1 with
      [] -> ids2
    | hd::tl -> subtracts tl (subtract hd ids2)

let rec fv (pre_t, typ) =
  match pre_t with
    (* Pcdata _ -> [] *)
  | String _ -> []
  | Int _ -> []
  | True | False -> []
  (* | Pcdata _ -> [] *)
  | Attribute _ -> []
  | Label _ -> []
  | AttrGet (a, x) ->
      (fv a)
  | BinOp (binop, t1, t2) -> (fv t1)@(fv t2) 
  | Var id -> [id]
  | Fix (f, x, t) ->
      subtract x (subtract f (fv t))
  | App (t1, t2) -> (fv t1)@(fv t2)
  | If (t0, t1, t2) -> (fv t0)@(fv t1)@(fv t2)
  | Let (x, t1, t2) ->
      subtract x ((fv t1)@(fv t2))
  | Letrecs(fts, t) ->
      let fs = List.map fst fts in
      let ts = List.map snd fts in
      let fv_tslist = List.map fv ts in
      let fvs = match fv_tslist with
	  hd::tl -> List.fold_left (@) hd tl in
	subtracts fs (fvs@(fv t))
  | Leaf t -> fv t
  | Node (t_label, t_attr, t_treelist)-> 
      (fv t_label)@(fv t_attr)@(fv t_treelist)
  | Nil -> []
  | Cons (t_hd, t_tl) -> (fv t_hd)@(fv t_tl)
  | Match (typ, id, matchs) ->
      id::(fv_matchs matchs)
and fv_matchs matchs =
  match matchs with
      [] -> []
    | (p, t)::tl ->
	(subtracts (fv_pattern p) (fv t))@(fv_matchs tl)
and fv_pattern p =
  match p with
      LeafPattern p -> fv_pattern p
    | NodePattern (p1, p2, p3) ->
	(fv_pattern p1)@(fv_pattern p2)@(fv_pattern p3)
    | NilPattern -> []
    | ConsPattern (p1, p2) -> (fv_pattern p1)@(fv_pattern p2)
    | IdentPattern id -> [id]
    | StringPattern str -> []
    | LabelPattern l -> []

let rec conv_typ typ
    = match typ with
	Typing.TString -> TString
      | Typing.TInt -> TInt
      | Typing.TBool -> TBool
(*      | Typing.TPcdata -> TPcdata *)
      | Typing.TLabel -> TLabel
      | Typing.TAttribute -> TAttribute
      | Typing.TFun(tdom, tcod) -> TFun(conv_typ tdom, conv_typ tcod)
      | Typing.TTree -> TTree (new_uvar ())
      | Typing.TTreelist -> TTreelist (new_uvar ())
      | Typing.TVar { contents = None } -> TTree (new_uvar ())
      | Typing.TVar { contents = Some t' } -> conv_typ t'
      (* Tuple, TupleGet *)
      | _ -> failwith "alpha.conv_type"

let rec alpha map t =
  let typ = conv_typ (snd t) in
    match (fst t) with
      (* Typing.Pcdata pc -> (Pcdata pc), typ *)
      | Typing.String s -> (String s), typ
      | Typing.Int n -> (Int n), typ
      | Typing.True -> True, typ
      | Typing.False -> False, typ
      (* | Typing.Pcdata pc -> (Pcdata pc), typ *)
      | Typing.Attribute a -> (Attribute a), typ
      | Typing.Label l -> (Label l), typ
      | Typing.AttrGet(a, x) ->
	  let a' = alpha map a in
	  let newx = new_id x in
	    (AttrGet (a', newx)), typ
      | Typing.BinOp(binop, t1, t2) ->
	  let conv_binop = function
	      Typing.Eq -> Eq
	    | Typing.Cat -> Cat
	    | Typing.Sub -> Sub
	    | Typing.ListCat -> ListCat
	    | Typing.Semi -> Semi
	    | Typing.Mod -> Mod
	  in 
	  BinOp((conv_binop binop), (alpha map t1), (alpha map t2)), typ
      | Typing.Var x -> (Var (List.assoc x map)), typ
      | Typing.Fix(f, x, t') ->
	  let newf = new_id f in
	  let newx = new_id x in
	    Fix(newf, newx, (alpha ((f, newf)::(x, newx)::map) t')), typ
      | Typing.App(t1, t2) -> App(alpha map t1, alpha map t2), typ
      | Typing.If(t0, t1, t2) ->
	  If(alpha map t0, alpha map t1, alpha map t2), typ
      | Typing.Let(x, t1, t2) ->
	  let newx = new_id x in
	    Let(newx, alpha map t1, alpha ((x, newx)::map) t2), typ
      | Typing.Letrecs(fts, t) ->
	  let rec map_letrecs map fts =
	    (match fts with
		 (fn, tn)::tl ->
		   let newfn = new_id fn in
		     (fn, newfn)::(map_letrecs map tl)
	       | [] -> map)
	  in
	  let map' = map_letrecs map fts in
	  let rec alpha_letrecs fts =
	    (match fts with
		 (fn, tn)::tl ->
		   ((List.assoc fn map'), (alpha map' tn))::(alpha_letrecs tl)
	       | [] -> [])
	  in
	  let fts' = alpha_letrecs fts in
	  let t' = alpha map' t in
	    Letrecs(fts', t'), typ
	    
(* 
     | Typing.Letrecs(fts, t) ->
	  let rec alpha_letrec map fts =
	    match fts with
		(fn, tn)::tl ->
		  let newfn = new_id fn in
		  let tn' = alpha map tn in
		    (newfn, tn')::(alpha_letrec ((fn, newfn)::map) tl)
	      | [] -> []
	  in
	  let fts' = alpha_letrec map fts in
	  let t' = alpha map t in
	    Letrecs(fts', t'), typ
*) 
      | Typing.Leaf t -> Leaf (alpha map t), typ
      | Typing.Node(t1, t2, t3) ->
	  Node(alpha map t1, alpha map t2, alpha map t3), typ
      | Typing.Nil -> Nil, typ
      | Typing.Cons(t1, t2) -> Cons(alpha map t1, alpha map t2), typ
      | Typing.Match(t, matchs) ->
	  let t' = alpha map t in
	  let temp, typ_match = t' in
	  let matchs' = List.map (alpha_one_match map) matchs in
	    (match (fst t') with
		Var x ->
		  Match(typ_match, x, matchs'), typ
	      | _ ->
		  let new_ident = new_id "dummyx" in
		    (App(
		       ((Fix(new_id "dummyf", new_ident,
			     (Match(typ_match, new_ident, matchs'), typ))), TFun((snd t'), typ)),
			t'))
		      , typ)



and alpha_one_match map (pattern, t) =
  let rec conv_p p map =
      match p with
	  Typing.LeafPattern p -> LeafPattern (conv_p p map)
	| Typing.NodePattern (p1, p2, p3) ->
	    NodePattern(conv_p p1 map, conv_p p2 map, conv_p p3 map)
	| Typing.NilPattern -> NilPattern
	| Typing.ConsPattern (p1, p2) ->
	    ConsPattern(conv_p p1 map, conv_p p2 map)
	| Typing.IdentPattern id -> IdentPattern (List.assoc id map)
	| Typing.StringPattern s -> StringPattern s
	| Typing.LabelPattern l ->  LabelPattern l
  in
  let rec make_pattern_map p=
    match p with
	Typing.LeafPattern p -> 
	  make_pattern_map p
      | Typing.NodePattern (p1, p2, p3) ->
	  (make_pattern_map p1)@(make_pattern_map p2)@(make_pattern_map p3)
      | Typing.NilPattern ->
	  []
      | Typing.ConsPattern (p1, p2) ->
	  (make_pattern_map p1)@(make_pattern_map p2)
      | Typing.StringPattern s -> 
	  []
      |	Typing.IdentPattern id -> 
	  [id, (new_id id)]
      | Typing.LabelPattern l -> 
	  []
  in 
  let map' = (make_pattern_map pattern)@map in
  let pattern' = conv_p pattern map' in
  let t' = alpha map' t in
    pattern', t'


let rec use_of_typ typ =
  match typ with
      TString | TInt | TBool | TLabel | TAttribute | TFun (_, _)
	-> UOmega
    | TTree u -> u
    | TTreelist u -> u
	
let rec make_pattern_env p typ =
  let use = use_of_typ typ in
  match p with
      LeafPattern p ->
	make_pattern_env p TString
    | NodePattern(p1, p2, p3) ->
	(make_pattern_env p1 TLabel)@(make_pattern_env p2 TAttribute)@(make_pattern_env p3 (TTreelist use))
    | NilPattern ->
	[]
    | ConsPattern (p1, p2) ->
	(make_pattern_env p1 (TTree use))@(make_pattern_env p2 (TTreelist use))
    | StringPattern s ->
	[]
    | IdentPattern id ->
	[id, typ]
    | LabelPattern l -> 
	[]

let rec id_of_pattern p =
  match p with
      LeafPattern p ->
	id_of_pattern p
    | NodePattern(p1, p2, p3) ->
	(id_of_pattern p1)@(id_of_pattern p2)@(id_of_pattern p3)
    | NilPattern ->
	[]
    | ConsPattern (p1, p2) ->
	(id_of_pattern p1)@(id_of_pattern p2)
    | StringPattern s ->
	[]
    | IdentPattern id ->
	[id]
    | LabelPattern l -> 
	[]
