type typ =
    Bottom
  | Top
  | TString
  | TInt
  | TBool
  | TLabel
  | TAttribute
  | TFun of typ * typ
  | TTree of (Alpha.label * typ) list * typ * typ
  | TList of typ
  | TTuple of typ list
  | TVar of int list
  | TRec of int list * typ

type constr = typ * typ

(* Convert Alpha.typ to typ : Alpha.typ -> typ *)
let rec conv_typ typ = match typ with
    Alpha.TString -> TString
  | Alpha.TInt -> TInt
  | Alpha.TBool -> TBool
  | Alpha.TLabel -> TLabel
  | Alpha.TAttribute -> TAttribute
  | Alpha.TFun(typ1, typ2) -> TFun(conv_typ typ1, conv_typ typ2)
  | Alpha.TTree _ -> TTree ([], Bottom, Bottom)
  | Alpha.TTreelist _ -> TList (TTree ([], Bottom, Bottom))
  | _ -> failwith "Cannot convert Alpha.typ to Constr_gen.typ"

(* Generate constraints : Typeinf.t -> constr list *)
let gen_constr t =
  let rec gen_constr' (t, typ_t) pos =
    match t with
	Typeinf.String s ->
	  [TString, TVar pos]
      | Typeinf.Int n ->
	  [TInt, TVar pos]
      | Typeinf.Label l ->
	  [TLabel, TVar pos]
      | Typeinf.Attribute a ->
	  [TAttribute, TVar pos]
      | Typeinf.AttrGet (a, x) ->
	  [TVar (0::pos), TVar pos]
      | Typeinf.True ->
	  [TBool, TVar pos]
      | Typeinf.False ->
	  [TBool, TVar pos]
      | Typeinf.Var x ->
	  [TVar [x.Alpha.id; -1], TVar pos]
      | Typeinf.BinOp(binop, (_,typ_t1 as t1), (_,typ_t2 as t2)) ->
	  begin
	    match binop with
		Typeinf.Eq ->
		  [(TBool, TVar pos);
		   (TVar (0::pos), conv_typ typ_t1);
		   (TVar (1::pos), conv_typ typ_t2)]
		  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
	      | Typeinf.Sub ->
		  [(TVar (0::pos), TVar pos);
		   (TVar (1::pos), TVar pos)]
		  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
	      | Typeinf.Cat ->
		  [(TVar (0::pos), TVar pos);
		   (TVar (1::pos), TVar pos)]
		  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
	      | Typeinf.ListCat ->
		  [(TVar (0::pos), TVar pos);
		   (TVar (1::pos), TVar pos)]
		  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
	      | Typeinf.Semi -> gen_constr' t2 (1::pos)
	      | Typeinf.Mod ->
		  [(TVar (0::pos), TVar pos);
		   (TVar (1::pos), TVar pos)]
		  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
	  end
      | Typeinf.App(t1, t2) ->
	  [(TVar (0::pos), TFun(TVar (-1::0::pos), TVar pos));
	   (TVar (1::pos), TVar (-1::0::pos))]
	  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
      | Typeinf.Fix(f, (x,_), t) ->
	  [(TFun(TVar [x.Alpha.id; -1], TVar (0::pos)), TVar pos);
	   (TFun(TVar [x.Alpha.id; -1], TVar (0::pos)), TVar [f.Alpha.id; -1])]
	  @ gen_constr' t (0::pos)
      | Typeinf.Let((x,_), t1, t2) ->
	  [(TVar (0::pos), TVar [x.Alpha.id;-1]);
	   (TVar (1::pos), TVar pos)]
	  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
      | Typeinf.Letrecs(fts, t) ->
	  let rec f fts n =
	    match fts with
		[] -> []
	      | (x,t)::fts ->
		  (TVar (n::pos), TVar [x.Alpha.id;-1])
		  :: gen_constr' t (n::pos) @ f fts (n+1)
	  in
	    (TVar (0::pos), TVar pos) :: f fts 1 @ gen_constr' t (0::pos)
      | Typeinf.If(t0, t1, t2) ->
	  [TVar (0::pos), TBool;
	   TVar (1::pos), TVar pos;
	   TVar (2::pos), TVar pos]
	  @ gen_constr' t0 (0::pos) @ gen_constr' t1 (1::pos) @ gen_constr' t2 (2::pos)
      | Typeinf.Leaf(_, t) ->
	  (TTree([], TVar (0::pos), Bottom), TVar pos)
	  :: gen_constr' t (0::pos)
      | Typeinf.Node(_, (Typeinf.Label l,typ), t2, t3) ->
	  [TVar (1::pos), TAttribute;
	   TTree ([(l,TVar(2::pos))], Bottom, Bottom), TVar pos]
	  @ gen_constr' t2 (1::pos) @ gen_constr' t3 (2::pos)
      | Typeinf.Node(_, t1, t2, t3) ->
	  [TVar (0::pos), TLabel;
	   TVar (1::pos), TAttribute]
	  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos) @ gen_constr' t3 (2::pos)
      | Typeinf.Cons(_, t1, t2) ->
	  [(TList (TVar (0::pos)), TVar pos);
	   (TVar (1::pos), TVar pos)]
	  @ gen_constr' t1 (0::pos) @ gen_constr' t2 (1::pos)
      | Typeinf.Nil _ -> [TList (TTree ([], Bottom, Bottom)), TVar pos]
      | Typeinf.Match(typ, y, matchs) ->
	  let rec f matchs n =
	    match matchs with
		[] -> []
	      | (_,t)::matchs ->
		  (TVar (n::pos), TVar pos)
		  :: gen_constr' t (n::pos) @ f matchs (n+1)
	  in
	  let rec g p =
	    match p with
		Alpha.LeafPattern p ->
(*
		  let f typ = TTree ([], typ, Top) in
		    List.map f (g p)
*)
		  let f typ = TTree ([], Bottom, Top) in
		    List.map f (g p)
	      | Alpha.NodePattern(pl,pa,ptl) ->
		  begin
		    match pl with
			Alpha.LabelPattern l ->
			  let f t = TTree ([(l,t)], Top, Top) in
			    List.map f (g ptl)
		      | _ ->
			  let f t = TTree ([], Top, t) in
			    List.map f (g ptl)
		  end
	      | Alpha.NilPattern -> [TList (TTree ([], Top, Top))]
	      | Alpha.ConsPattern(pt, ptl) ->
		  let f typ = TList typ in
		    List.map f (g pt) @ g ptl
	      | Alpha.IdentPattern x -> [TVar [x.Alpha.id;-1]]
	      | Alpha.StringPattern _ -> [TString]
	      | Alpha.LabelPattern _ -> [TLabel]
	  in
	  let h typ = (TVar [y.Alpha.id;-1], typ) in
	    f matchs 0 @ List.map h (List.flatten (List.map (fun(p,_)->g p) matchs))
      | Typeinf.LetS2M_t(id, t) ->
	  (TVar (0::pos), TVar pos) :: gen_constr' t (0::pos)
      | Typeinf.LetS2M_tl(id, t) ->
	  (TVar (0::pos), TVar pos) :: gen_constr' t (0::pos)
      | Typeinf.M2S_t t ->
	  (TVar (0::pos), TVar pos) :: gen_constr' t (0::pos)
      | Typeinf.M2S_tl tl ->
	  (TVar (0::pos), TVar pos) :: gen_constr' tl (0::pos)
      | Typeinf.Copy t -> [TVar [t.Alpha.id;-1], TVar pos]
      | Typeinf.Skip(_, t) ->
	  (TVar (0::pos), TVar pos) :: gen_constr' t (0::pos)
(*
      | Typeinf.StringOfInt t ->
	  [TString, TVar pos]
      | Typeinf.Primitive _ -> [TFun (TVar (0::pos), TVar (0::pos)), TVar pos]
*)
      | _ -> failwith "Cannot occur : constr_gen"
  in
  let toptyp = TFun(TTree([], Bottom, Bottom), TTree([], Bottom, Bottom)) in
    (TVar [], toptyp) :: gen_constr' t []

(* Print type : typ -> () *)
let rec print_typ typ =
  match typ with
      Bottom -> print_string "Bottom"
    | Top -> print_string "Top"
    | TString -> print_string "TString"
    | TInt -> print_string "TInt"
    | TBool -> print_string "TBool"
    | TLabel -> print_string "TLabel"
    | TAttribute -> print_string "TAttribute"
    | TFun(typ1, typ2) ->
	print_string "(";
	print_typ typ1;
	print_string " -> ";
	print_typ typ2;
	print_string ")"
    | TTree(lts, tl, tw) -> 
	let f (l,typ) =
	  print_string l;
	  print_string ":";
	  print_typ typ;
	  print_string ", "
	in
	  print_string "<"; List.iter f lts;
	  print_string "Leaf:"; print_typ tl;
	  print_string ">^"; print_typ tw
    | TList typ -> print_typ typ; print_string " list"
    | TTuple typs ->
	let rec f typs = match typs with
	    typ::[] -> print_typ typ
	  | typ::typs -> print_typ typ; print_string ", "
	  | [] -> failwith "Cannot occur : print_typ"
	in
	  print_string "(";
	  f typs;
	  print_string ")"
    | TVar ns ->
	begin
	  match ns with
	      [n; -1] -> print_string "beta"; print_int n
	    | [n; -2] -> print_string "alphaS"; print_int n
	    | -1::ns -> print_string "gamma"; List.iter print_int ns
	    | _ -> print_string "alpha"; List.iter print_int ns
	end
    | TRec(ns, typ) ->
	print_string "mu[";
	print_typ (TVar ns);
	print_string "].(";
	print_typ typ;
	print_string ")"

(* Print constraints : constr list -> () *)
let rec print_constrs constrs =
  match constrs with
      [] -> ()
    | (typ1,typ2)::constrs ->
	print_typ typ1;
	print_string " <: ";
	print_typ typ2;
	print_newline ();
	print_constrs constrs
