open Constr_gen

(* Union xs and ys as set : 'a list -> 'a list -> 'a list *)
let union xs ys =
  let rec add xs ys =
    match ys with
	[] -> xs
      | y::ys -> if List.mem y xs then add xs ys else add (y::xs) ys
  in
    add (add [] xs) ys

(* Get reflective closure : typ -> constr list -> typ list *)
let refl_clos typ constrs =
  let rec refl_clos' typ typs n =
    let f typ = List.filter (fun (typ1,typ2) -> typ = typ1) constrs in
    let next = List.map snd (List.flatten (List.map f typs)) in
    let typs = union typs next in
    let m = List.length typs in
      if n = m then typs else refl_clos' typ typs m
  in
    refl_clos' typ [typ] 1

(* Get all types in constraints : constr list -> typ list *)
let get_typs constrs =
  let rec get_sub typ =
    let subs =
      match typ with
	  TFun(typ1, typ2) -> get_sub typ1 @ get_sub typ2
	| TTree(typs, typ1, typ2) ->
	    let f (_, typ) = get_sub typ in
	      List.flatten (List.map f typs) @ get_sub typ1 @ get_sub typ2
	| TList typ1 -> get_sub typ1
	| TTuple typs -> List.flatten (List.map get_sub typs)
	| _ -> []
    in
       typ :: subs
  in
  let f (typ1, typ2) =
    get_sub typ1 @ get_sub typ2
  in
    union [] (List.flatten (List.map f constrs))

(* Get type of label : typ -> label -> Alpha.label list -> constr list *)
let get_label_typ typ l =
  match typ with
      TTree(lts, _, typ) ->
	(try List.assoc l lts with Not_found -> typ)
    | _ -> failwith "Cannot occur : get_label_typ"

(* Solve constraints (step 2) : constr list -> constr list *)
let solv_constr1 constrs =
  let rec solv_constr1' constrs n =
    let solv_constr1'' typ =
      let f typ' =
	match typ, typ' with
	    TFun(typ1,typ2), TFun(typ1',typ2') -> [(typ1',typ1); (typ2,typ2')]
	  | TTree(lts, typ1, typ2), TTree(lts', typ1', typ2') ->
	      let labels = union (List.map fst lts) (List.map fst lts') in
	      let f l = (get_label_typ typ l, get_label_typ typ' l) in
		(typ1,typ1') :: (typ2,typ2') :: List.map f labels
	  | TList typ, TList typ' -> [typ, typ']
	  | TTuple typs, TTuple typs' -> List.combine typs typs'
	  | TFun(typ1,typ2), Bottom -> [Top,typ1; typ2,Bottom]
	  | TTree(lts, typ1, typ2), Bottom ->
	      let f (l, typ) = (typ, Bottom) in
		(typ1,Bottom) :: (typ2,Bottom) :: List.map f lts
	  | TList typ, Bottom -> [typ, Bottom]
	  | TTuple typs, Bottom ->
	      let f typ = (typ, Bottom) in
		List.map f typs
	  | _ -> []
      in
      let rclos = List.filter ((<>) typ) (refl_clos typ constrs) in
	List.flatten (List.map f rclos)
    in
    let typs = List.map fst constrs in
    let constrs' = List.flatten (List.map solv_constr1'' typs) in
    let constrs = union constrs constrs' in
    let m = List.length constrs in
      if n = m then constrs else solv_constr1' constrs m
  in
  let constrs = union [] constrs in
    solv_constr1' constrs (List.length constrs)

(* Solve constraints (step 3) : constr list -> constr list *)
let solv_constr2 constrs =
  let get_constrs cnstrs =
    let f ((typ1,_), (typ2,_)) = (typ1, typ2) in
      List.map f cnstrs
  in
  let new_var =
    let counter = ref 0 in
      fun () -> incr counter; TVar [!counter; -2]
  in
  let rec solv_cnstr2' cnstrs =
    let rec solv_cnstr2'' cnstrs (var,typs) =
      let exist typs =
	let vertexs = List.map fst cnstrs in
	try
	  Some (fst (List.find (fun (var,typs') -> typs = typs') vertexs))
	with
	    Not_found -> None
      in
      let rec new_typ_cnstrs var typs =
	if List.mem Bottom typs then (Bottom, [])
	else
	  match List.hd typs with
	      TFun _ ->
		let newvar1 = new_var () in
		let newvar2 = new_var () in
		let newtyp = TFun(newvar1, newvar2) in
		let newcnstrs =
		  let f typ =
		    match typ with
			TFun(typ1, typ2) -> [((typ1,[]), (newvar1,[]));
					     ((newvar2,[]), (typ2,[]))]
		      | _ -> failwith "Cannot occur : solv_cnstr2''(TFun)"
		  in
		    List.flatten (List.map f typs)
		in
		  (newtyp, newcnstrs)
	    | TTree _ ->
		let labels =
		  let f typ =
		    match typ with
			TTree(lts, _, _) -> List.map fst lts
		      | _ -> failwith "Cannot occur : solv_cnstr2''(TTree)"
		  in
		    union [] (List.flatten (List.map f typs))
		in
		let newvars = List.map (fun _ -> new_var ()) labels in
		let newlvs = List.combine labels newvars in
		let newvar1 = new_var () in
		let newvar2 = new_var () in
		let newtyp = TTree (newlvs, newvar1, newvar2) in
		let newcnstrs =
		  let f typ =
		    match typ with
			TTree(_, typ1, typ2) ->
			  let f var l = ((var,[]), (get_label_typ typ l,[])) in
			    ((newvar1,[]),(typ1,[])) :: ((newvar2,[]),(typ2,[])) ::
			      List.map2 f newvars labels
		      | _ -> failwith "Cannot occur : solv_cnstr2''(newcnstrs)"
		  in
		    List.flatten (List.map f typs)
		in
		  (newtyp, newcnstrs)
	    | TList _ ->
		let newvar = new_var () in
		let newtyp = TList newvar in
		let newcnstrs =
		  let f typ =
		    match typ with
			TList typ -> ((newvar,[]), (typ,[]))
		      | _ -> failwith "Cannot occur : solv_cnstr2''(TList)"
		  in
		    List.map f typs
		in
		  (newtyp, newcnstrs)
	    | TTuple typs' ->
		let newvars = List.map (fun _ -> new_var ()) typs' in
		let newtyp = TTuple newvars in
		let newcnstrs =
		  let f var typ = ((var,[]), (typ,[])) in
		  let g typ =
		    match typ with
			TTuple typs -> typs
		      | _ -> failwith "Cannot occur : solv_cnstr2''(TTuple)"
		  in
		    List.flatten (List.map (fun typs -> List.map2 f newvars typs) (List.map g typs))
		in
		  (newtyp, newcnstrs)
	    | _ -> failwith "Cannot occur : solv_cnstr2''"
      in
      let typs = List.filter ((<>) Top) typs in
      let (newtyp, newcnstrs) =
	match exist typs with
	    Some var -> (var, [])
	  | None -> new_typ_cnstrs var typs
      in
      let f cnstr =
	match cnstr with ((var',typs'), vts) ->
	  if var = var' then ((newtyp,typs), vts) else cnstr
      in
	((var,[]),(newtyp,typs)) :: List.map f cnstrs @ newcnstrs
    in
    let typs = get_typs (get_constrs cnstrs)
    in
    let vars = List.filter (function TVar _ -> true | _ -> false) typs in
    let vars =
      let f var = List.filter ((=) (var,[])) (List.map fst cnstrs) in
      let f var = (var, f var) in
      let vtss = List.filter (fun (var,typs) -> List.length typs > 1) (List.map f vars) in
	List.map fst vtss
    in
    let vtss =
      let f = List.filter (function TVar _ -> false | _ -> true) in
      let f var = (var, union [] (f (refl_clos var (get_constrs cnstrs)))) in
	List.map f vars
    in
    let vtss = List.filter (fun (_,typs) -> List.length typs > 1) vtss in
    let cnstr () = List.fold_left solv_cnstr2'' cnstrs vtss in
      if vtss = [] then cnstrs else solv_cnstr2' (cnstr ())
  in
  let f (typ1, typ2) = ((typ1,[]), (typ2,[])) in
    get_constrs (solv_cnstr2' (List.map f constrs))

(* Remove abundant constraints : constr list -> constr list *)
let reduce_constr constrs =
  let constrs = union [] constrs in
  let f (typ, _) = match typ with TVar _ -> true | _ -> false in
  let constrs = List.filter f constrs in
  let f (_, typ) = typ <> Top in
  let constrs = List.filter f constrs in
    List.rev constrs

(* Simplify a type : typ -> typ *)
let simplify typ =
  let rec simplify' typ =
    match typ with
	TFun(typ1, typ2) ->
	  let (typ1, nss1) = simplify' typ1 in
	  let (typ2, nss2) = simplify' typ2 in
	    (TFun(typ1, typ2), nss1 @ nss2)
      | TTree(lts, typ1, typ2) ->
	  let f (l, typ) =
	    let (typ, nss) = simplify' typ in
	      ((l, typ), nss)
	  in
	  let ltnsss = List.map f lts in
	  let lts = List.map fst ltnsss in
	  let nss = List.flatten (List.map snd ltnsss) in
	  let (typ1, nss1) = simplify' typ1 in
	  let (typ2, nss2) = simplify' typ2 in
	    (TTree(lts, typ1, typ2), nss @ nss1 @ nss2)
      | TList typ ->
	  let (typ, nss) = simplify' typ in
	    (TList typ, nss)
      | TTuple typs ->
	  let tnsss = List.map simplify' typs in
	  let typs = List.map fst tnsss in
	  let nss = List.flatten (List.map snd tnsss) in
	    (TTuple typs, nss)
      | TVar ns -> (typ, [ns])
      | TRec(ns, typ) ->
	  let (typ, nss) = simplify' typ in
	  let nss = List.filter ((=) ns) nss in
	  let typ = if List.mem ns nss then TRec(ns, typ) else typ in
	    (typ, nss)
      | _ -> (typ, [])
  in
  let (typ, _) = simplify' typ in
    typ

(* Specify a type : int list -> cnstr list -> rectyp *)
let get_typ pos constrs =
  let rec get_typ' typ nss =
    let get_var var =
      let f (typ1,typ2) = typ1 = var in
      let next = List.filter f constrs in
      let next = List.map snd next in
      let f typ = match typ with TVar _ -> false | _ -> true in
      let next = if List.length next > 1 then List.filter f (refl_clos var constrs) else next in
	match next with
	    [] -> Top
	  | [typ] -> typ
	  | _ -> failwith "Cannot occur : solv_constr";
    in
      match typ with
	| TFun(typ1, typ2) -> TFun(get_typ' typ1 nss, get_typ' typ2 nss)
	| TTree(lts, typ1, typ2) ->
	    let f (l, typ) = (l, get_typ' typ nss) in
	      TTree(List.map f lts, get_typ' typ1 nss, get_typ' typ2 nss)
	| TList typ -> TList (get_typ' typ nss)
	| TTuple typs ->
	    let f typ = get_typ' typ nss in
	      TTuple (List.map f typs)
	| TVar ns ->
	    if List.mem ns nss then typ else TRec (ns, get_typ' (get_var typ) (ns::nss))
	| _ -> typ
  in
    simplify (get_typ' (TVar pos) [])
