
open Format
open Translation

let rec generate_t chan formatter t =
  begin
    match t with
    | Leaf (Alpha.UOne, t) -> fprintf chan "(RuntimeType.Pcdata(%a))" (generate_t chan) t
    | Leaf (Alpha.UOmega, t) -> fprintf chan "(RuntimeType.Leaf(%a))" (generate_t chan) t
    | String s -> fprintf chan (* "\"%s\"" *) "(\"%s\")" s
      | Int n -> fprintf chan "%d" n
      | True -> fprintf chan "true"
      | False -> fprintf chan "false"
      | Var x -> fprintf chan "%s" x.Alpha.origin
      | Tuple ts ->
	  let rec print_ts formatter ts =
	    match ts with
		[hd] -> fprintf chan "%a" (generate_t chan) hd
	      | hd::tl -> fprintf chan "%a, %a" (generate_t chan) hd print_ts tl
	  in
	    fprintf chan "(%a)" print_ts ts
      | TupleGet(nelem, t, n) ->
	  let newid = Alpha.new_id "_" in
	  let rec print_tupleget_pattern n' =
	    match n' with
	      0 -> if n = n' then fprintf chan "%s" newid.Alpha.origin else fprintf chan "_"
	    | _ -> 
		(if n = nelem - n' then fprintf chan "%s, " newid.Alpha.origin else fprintf chan "_, ");
		print_tupleget_pattern (n' - 1)
	  in
	    fprintf chan "(match %a with " (generate_t chan) t;
	    print_tupleget_pattern (nelem - 1);
	  fprintf chan " -> %s)" newid.Alpha.origin
      | BinOp(b, t1, t2) ->
	  begin
	    match b with
	      Typeinf.Cat | Typeinf.ListCat -> fprintf chan "(%a;%a)" (generate_t chan) t1 (generate_t chan) t2
	    | _ -> fprintf chan "(%a %a %a)" (generate_t chan) t1 (Typeinf.print_binop) b  (generate_t chan) t2
	  end
      | Assert t ->
	  fprintf chan "(assert %a)" (generate_t chan) t
      | If(t, t1, t2) ->
	  fprintf chan "(if @[%a@] then @[%a@] else @[%a@])" (generate_t chan) t (generate_t chan) t1 (generate_t chan) t2
(*
      | Fix(fs, ts) ->
	  let rec print_fix fs ts =
	    match fs, ts with
		[hd], [t] -> fprintf chan "%s = @[%a@] in@." hd (generate_t chan) t
	      | hd::tl, t_hd::t_tl ->
		  fprintf chan "%s = @[%a@] and " hd (generate_t chan) t_hd;
		  print_fix tl t_tl
	  in
	  let rec print_tuple formatter = 
	    function [hd] -> fprintf chan "%s" hd | hd::tl -> fprintf chan "%s, " hd; print_tuple formatter tl 
	  in
	    begin
	      (* fprintf chan "(let rec %s = %a in %s)" f (generate_t chan) t f *)
	      fprintf chan "let rec ";
	      print_fix (List.map (fun x -> x.Alpha.origin) fs) ts;
	      fprintf chan "(%a)" print_tuple (List.map (fun x -> x.Alpha.origin) fs)
	    end
*)
      | Fix([f], [t]) ->
	  fprintf chan "(fun %s -> @[%a@])" f.Alpha.origin (generate_t chan) t
      | Lam((x, _), t) ->
	  fprintf chan "(fun %s -> @[%a@])" x.Alpha.origin (generate_t chan) t
      | App(Lam((x, _), t2), t1) | Let((x, _), t1, t2) ->
	  fprintf chan "(let %s = @[%a@] in@. @[%a@])" x.Alpha.origin (generate_t chan) t1 (generate_t chan) t2
      | App(t1, t2) ->
	  fprintf chan "(%a %a)" (generate_t chan) t1 (generate_t chan) t2
      | Write t ->
	  fprintf chan "(Runtime.write %a)" (generate_t chan) t
      (*| Leaf t ->
	  fprintf chan "(Leaf %a)" (generate_t chan) t *)
      | Node(t_l, t_a, t) ->
	  fprintf chan "(RuntimeType.Node(%a, %a, %a))" (generate_t chan) t_l (generate_t chan) t_a (generate_t chan) t
      | StartTag(t_l, t_attr) ->
	  (* fprintf chan "(\"<%a %a>\")" (generate_t chan) t_l (generate_t chan) t_attr *)
	  fprintf chan "(RuntimeType.StartTag(%a,%a))" (generate_t chan) t_l (generate_t chan) t_attr
      | CloseTag t_l ->
	  (* fprintf chan "(\"</%a>\")" (generate_t chan) t_l *)
	  fprintf chan "(RuntimeType.EndTag(%a))" (generate_t chan) t_l
      | Seq [] ->
	  fprintf chan "()"
      | Seq [hd] ->
	  fprintf chan "(%a)" (generate_t chan) hd
      | Seq (hd::tl) ->
	  fprintf chan "(%a;%a)" (generate_t chan) hd (generate_t chan) (Seq tl)
      | Attr attribute ->
	  fprintf chan "[%a]" (generate_attrs chan) attribute
      | AttrGet(t, x) ->
	  fprintf chan "(List.assoc \"%s\" %a)" x.Alpha.origin (generate_t chan) t
      | Unit ->
	  fprintf chan "()"
      | Nil ->
	  fprintf chan "[]"
      | Label l ->
	  fprintf chan "\"%s\"" l
      | Cons(t1, t2) ->
	  fprintf chan "(%a::%a)" (generate_t chan) t1 (generate_t chan) t2
      | Push t ->
	  fprintf chan "(Runtime.push %a)" (generate_t chan) t
      | Pop ->
	  fprintf chan "(Runtime.pop ())"
      | Read ->
	  fprintf chan "(Runtime.read_token ())"
      | Peek ->
	  fprintf chan "(Runtime.peek_token ())"
      | Match(t, pats) ->
	  fprintf chan "(match %a with @[%a@])" (generate_t chan) t (generate_pats chan) pats      
      | Letrecs(defs, t) ->
	  let rec print_letrec = function
	      [f, t] -> fprintf chan "%s = @[%a@] in" f.Alpha.origin (generate_t chan) t
	    | (f, t)::tl ->
		fprintf chan "%s = @[%a@] and " f.Alpha.origin (generate_t chan) t;
		print_letrec tl
	  in
	  fprintf chan "let rec ";
	  print_letrec defs;
	  fprintf chan "@[%a@]" (generate_t chan) t
(*
      | Comment s ->
	  fprintf chan "(* %s *)" s
*)
  end
and generate_attr chan formatter = function
    k, (String s) -> fprintf chan "(\"%s\", \"%s\")" k s
  | k, v -> fprintf chan "(\"%s\", %a)" k (generate_t chan) v
and generate_attrs chan formatter = function
    [] -> ()
  | [hd] -> (* (generate_attr chan formatter hd) *) ()
  | hd::tl -> (* fprintf chan "%a; %a" (generate_attr chan) hd (generate_attrs chan) tl*) ()
(*
  | [k, v] -> fprintf chan "(\"%s\", %a)" k (generate_t chan) v
  | (k, v)::tl -> fprintf chan "(\"%s\", %a); %a" k (generate_t chan) v (generate_attr chan) tl
*)
and generate_pats chan formatter = function
    [] -> ()
  | [p, t] -> fprintf chan "%a -> @[%a@]" (generate_pattern chan) p (generate_t chan) t
  | (p, t)::tl -> fprintf chan "%a -> @[%a@] @[| %a@]" (generate_pattern chan) p (generate_t chan) t (generate_pats chan) tl
and generate_pattern chan formatter = function  
    NilPattern -> fprintf chan "[]"
  | ConsPattern(p1, p2) -> fprintf chan "(%a::%a)" (generate_pattern chan) p1 (generate_pattern chan) p2
  | NodePattern(p_l, p_a, p_children) ->
      fprintf chan "(Node(%a, %a, %a))" (generate_pattern chan) p_l (generate_pattern chan) p_a (generate_pattern chan) p_children
  | LeafPattern (Alpha.UOne, p) ->
      fprintf chan "RuntimeType.Pcdata %a" (generate_pattern chan) p
  | LeafPattern (Alpha.UOmega, p) ->
      fprintf chan "RuntimeType.Leaf %a" (generate_pattern chan) p
  | StringPattern s ->
      fprintf chan "%s" s
  | LabelPattern s ->
      fprintf chan "\"%s\"" s
  | StartTagPattern (p, p_a) ->
      fprintf chan "(RuntimeType.StartTag(%a, %a))" (generate_pattern chan) p (generate_pattern chan) p_a
  | EndTagPattern p ->
      fprintf chan "(RuntimeType.EndTag %a)" (generate_pattern chan) p
  | IdentPattern x ->
      fprintf chan "%s" x.Alpha.origin
  | AttributePattern a ->
      fprintf chan "[%a]" (generate_attribute chan) a
  | OrPattern(p1, p2) ->
      fprintf chan "(%a | %a)" (generate_pattern chan) p1 (generate_pattern chan) p2
and generate_attribute chan formatter = function
    [] -> ()
  | [k, v] -> fprintf chan "(%s, %s)" k v
  | (k, v)::tl -> fprintf chan "(%s, %s); %a" k v (generate_attribute chan) tl

let generate chan t =
  fprintf chan
"open RuntimeType
 open XmlLexer
 open Runtime

 let main filename =
  let in_channel =
   match filename with
     None -> stdin
   | Some f -> open_in f
  in
  let lexbuf = Lexing.from_channel in_channel in
  let _ = Runtime.lexbuf := lexbuf in
   %a

 let _ =
  (* Gc.set { (Gc.get ()) with Gc.verbose = 0x004 }; *)

  let start = Unix.gettimeofday () in
    if !Sys.interactive then
     ()
    else if Array.length Sys.argv = 2 then
    begin
     (main (Some Sys.argv.(1))) ();
     (* Printf.fprintf stderr \"%%f \" (Gc.allocated_bytes ()) *)
    end
    else
    begin
     (main None\n) ();
     (* Printf.fprintf stderr \"%%f \" (Gc.allocated_bytes ()) *)
    end;
    (* Printf.fprintf stderr \"%%f\\n\" ((Unix.gettimeofday ()) -. start) *)
    Printf.fprintf stderr \"\\n\""
       (generate_t chan) t
