
open Format

type ident = Typeinf.ident
type use = Typeinf.use
type label = Typeinf.label
type attribute = Typeinf.attribute

let new_id = Alpha.new_id

type t =
  | String of string
  | Int of int
  | Tuple of t list
  | TupleGet of int * t * int
  | True | False
  | Var of ident
  | BinOp of Typeinf.binop * t * t
  | Assert of t
  | If of t * t * t
  | Fix of ident list * t list
  | Letrecs of (ident * t) list * t
  (* | Letrecs of (ident * t) list *)
  | Lam of (ident * use) * t
  | App of t * t
  | Let of (ident * use) * t * t
  | Write of t
  (* | Leaf of t *)
  | Node of (* label *) t * t * t
  | Leaf of use * t
  | StartTag of (* label *) t * t
  | CloseTag of (* label *) t
  | Seq of t list
  | Attr of attribute
  | AttrGet of t * ident
  | Unit
  | Nil
  | Label of label
  | Cons of t * t
  | Push of t
  | Pop
  | Read
  | Peek
  | Match of t * ((pattern * t) list)
  (* | Comment of string *)
(*
  | MatchStart of t * ident (* label *) * ident (* attributes *) * t * ident (* pcdata *) * t
  | MatchEnd of t * ident (* label *) * t * t
  | MatchTree of t * ident (* PCDATA *) * t * ident (* label *) * ident (* attr list *) * ident (* tree list *) * t
  | MatchList of t * t * ident (* head *) * ident (* tail *) * t
*)
and pattern =
    NilPattern
  | ConsPattern of pattern * pattern
  | NodePattern of pattern * pattern * pattern
  | LeafPattern of use * pattern
  | StringPattern of string
  | LabelPattern of string
  | StartTagPattern of pattern * pattern
  | EndTagPattern of pattern
  | IdentPattern of ident
  | AttributePattern of (string * string) list
  | OrPattern of pattern * pattern

let rec unwind_let t = match t with
  Leaf (u, t) -> Leaf (u, (unwind_let t))
  | ((* Comment _ | *) Unit | Nil | Label _ | Pop | Read | Peek | Attr _ | True | False | Var _ | String _ | Int _) as t -> t
  | Tuple ts -> Tuple (List.map unwind_let ts)
  | TupleGet(n, t, n') -> TupleGet(n, unwind_let t, n')
  | BinOp(b, t1, t2) -> BinOp(b, unwind_let t1, unwind_let t2)
  | Assert t -> Assert (unwind_let t)
  | If(t, t1, t2) -> If(unwind_let t, unwind_let t1, unwind_let t2)
  | Fix(fs, ts) -> Fix(fs, List.map unwind_let ts)
  | Lam(x, t) -> Lam(x, unwind_let t)
  | App(Lam(x, t2), t1) -> 
      unwind_let (Let(x, unwind_let t1, unwind_let t2))
  | App(t1, t2) -> App(unwind_let t1, unwind_let t2)
  | Let(x1, Let(x2, t2, t3), t4) ->
      unwind_let (Let(x2, unwind_let t2, Let(x1, unwind_let t3, unwind_let t4)))
  | Let(x1, t1, t2) ->
      Let(x1, unwind_let t1, unwind_let t2)
(*
      let t1' = unwind_let t1 in
      let t2' = unwind_let t2 in
	begin
	  match t1' with
	      (* Let(x1, Let(x2, t3, t4), t2') *)
	      Let(x2, t3, t4) -> unwind_let (Let(x2, t3, Let(x1, t4, t2')))
	    | _ -> Let(x1, t1', t2')
	end
*)
  | Write t -> Write (unwind_let t)
  | Node(t, t1, t2) -> Node(unwind_let t, unwind_let t1, unwind_let t2)
  | StartTag(t1, t2) -> StartTag(unwind_let t1, unwind_let t2)
  | CloseTag t -> CloseTag(unwind_let t)
  (* | Seq ts -> Seq (List.map unwind_let ts) *)
  | Seq ts ->
      let rec unwind_seq = function 
	  [] -> [] 
	| (Seq ts')::tl -> (unwind_seq ts') @ (unwind_seq tl)
	| hd::tl -> hd::(unwind_seq tl)
      in
	Seq (List.map unwind_let (List.filter (function Unit | _ -> true) (unwind_seq ts)))
  | AttrGet(t, x) -> AttrGet(unwind_let t, x)
  | Cons(t1, t2) -> Cons(unwind_let t1, unwind_let t2)
  | Push t -> Push (unwind_let t)
  | Match(t, pats) -> Match(unwind_let t, List.map (fun (p, t) -> (p, unwind_let t)) pats)

(* [XX] assumes alpha conversion *)
(* returns [t1/x]t2 *)
let rec substitute x t1 t2 =
  match t2 with
      Var x' ->	if x = x' then t1 else t2
    | True | False | Int _ | String _ -> t2
    | Label l -> Label l
    | Fix(fs, ts) -> Fix(fs, List.map (substitute x t1) ts)
    | Lam((x', u), t) -> Lam((x', u), substitute x t1 t)
    | Let(x', t1', t2') -> Let(x', substitute x t1 t1', substitute x t1 t2')
    | App(t', t'') -> App(substitute x t1 t', substitute x t1 t'')
    | Write t -> Write (substitute x t1 t)
    (* | Leaf t -> Leaf (substitute x t1 t) *)
    | Node(l, t', t'') -> Node(l, substitute x t1 t', substitute x t1 t'')
    | StartTag(l, t) -> StartTag(l, substitute x t1 t)
    | Seq ts -> Seq (List.map (substitute x t1) ts)
    | Cons(t_hd, t_tl) -> Cons(substitute x t1 t_hd, substitute x t1 t_tl)
    | Push t -> Push (substitute x t1 t)
(*
    | MatchStart(t, l, attr, t', x, t'') ->
	MatchStart(substitute x t1 t, l, attr, substitute x t1 t', x, substitute x t1 t'')
    | MatchTree(t, x, t', l, attr, ts, t'') ->
	MatchTree(substitute x t1 t, x, substitute x t1 t', l, attr, ts, substitute x t1 t'')
    | MatchList(t, t', hd, tl, t'') -> MatchList(substitute x t1 t, substitute x t1 t', hd, tl, substitute x t1 t'')
    | MatchEnd(t, x, t', t'') -> MatchEnd(substitute x t1 t, x, substitute x t1 t', substitute x t1 t'')
*)
    | Match(t, pats) -> Match(substitute x t1 t, List.map (fun (p, t) -> (p, substitute x t1 t)) pats)
    | If(t, t', t'') -> If(substitute x t1 t, substitute x t1 t', substitute x t1 t'')
    | Assert t -> Assert (substitute x t1 t)
    | BinOp(b, t, t') -> BinOp(b, substitute x t1 t, substitute x t1 t')
    | Attr attr -> (* Attr (List.map (fun (l, t) -> (l, substitute x t1 t)) attr) *) Attr attr
    | AttrGet(t, l) -> AttrGet(substitute x t1 t, l)
    | Leaf (u, t) -> Leaf (u, (substitute x t1 t))
    | (CloseTag _|String _|Peek|Read|Pop|Nil|Unit) -> t2
    | TupleGet(n, t, n') -> TupleGet(n, substitute x t1 t, n')
    | Tuple ts -> Tuple (List.map (substitute x t1) ts)

(*
  returns
  - the result of translation,
  - a list of free variables that must be converted to unit and
  - a list of labels that should be read
*)
let rec translate_pattern p u =
  match p, u with
    Alpha.NilPattern, Alpha.UOmega -> NilPattern, [], []
  | Alpha.NilPattern, _ -> (* NilPattern *) (* IdentPattern "_" *) EndTagPattern(IdentPattern (new_id "_")), [], []
  | Alpha.ConsPattern(p1, p2), Alpha.UOmega -> 
      let p1', fv1, lab1 = translate_pattern p1 u in
      let p2', fv2, lab2 = translate_pattern p2 u in
	ConsPattern (p1', p2'), fv1 @ fv2, lab1
  (* | Alpha.ConsPattern(Alpha.One, p1, p2) -> IdentPattern "_", [], [] *)
  | Alpha.ConsPattern(p1, p2), Alpha.UOne -> 
      OrPattern(StartTagPattern (IdentPattern (new_id "_"), IdentPattern (new_id "_")), 
		LeafPattern (Alpha.UOne, IdentPattern (new_id "_"))),
      [], []
  | (Alpha.LeafPattern p), u -> 
      let p', fv, lab = translate_pattern p u in
	LeafPattern (u, p'), [], lab
  | (Alpha.StringPattern s), _ -> StringPattern s, [], []
  | (Alpha.IdentPattern x), Alpha.UOmega -> IdentPattern x, [x], []
  | (Alpha.IdentPattern x), _ -> IdentPattern x, [x], []
  | (Alpha.LabelPattern l), _ -> LabelPattern l, [], [`Lab l]
  | (Alpha.NodePattern(Alpha.IdentPattern l, p_a, p)), Alpha.UOmega ->
      (* let p_l', fv_l, lab_l = translate_pattern p_l in *)
      let p_a', fv_a, lab_a = translate_pattern p_a u in
      let p', fv, lab = translate_pattern p u in
	(* Only free variables in p should be converted to unit *)
      NodePattern(IdentPattern l, p_a', p'), fv, (`Var l)::lab
  | Alpha.NodePattern(Alpha.IdentPattern l, p_a, p), Alpha.UOne ->
      (* let p_l', fv_l, lab_l = translate_pattern p_l in *)
      let p_a', fv_a, lab_a = translate_pattern p_a u in
      let p', fv, lab = translate_pattern p u in
	StartTagPattern(IdentPattern l, p_a'), fv, (`Var l)::lab
  | Alpha.NodePattern(p_l, p_a, p), Alpha.UOmega ->
      let p_l', fv_l, lab_l = translate_pattern p_l Alpha.UOmega in
      let p_a', fv_a, lab_a = translate_pattern p_a Alpha.UOmega in
      let p', fv, lab = translate_pattern p u in
	(* Only free variables in p should be converted to unit *)
      NodePattern(p_l', p_a', p'), fv, lab_l @ lab
  | Alpha.NodePattern(p_l, p_a, p), Alpha.UOne ->
      let p_l', fv_l, lab_l = translate_pattern p_l Alpha.UOmega in
      let p_a', fv_a, lab_a = translate_pattern p_a Alpha.UOmega in
      let p', fv, lab = translate_pattern p u in
	StartTagPattern(p_l', p_a'), fv, lab_l @ lab

let rec translate = function
  | Typeinf.Leaf (Alpha.UPlus, t) -> Write (* (Pcdata (translate t)) *) (Leaf (Alpha.UOne, (translate (fst t))))
  | Typeinf.Leaf (Alpha.UOmega, t) -> Leaf (Alpha.UOmega, (translate (fst t)))
  | Typeinf.String s -> String s (* (Pcdata (String s)) *)
  | Typeinf.Int n -> Int n
(*
  | Typeinf.Tuple ts -> Tuple (List.map translate ts)
  | Typeinf.TupleGet(typ, t, n) ->
      begin
	match Typing.shorten typ with
	    Typeinf.TTuple l -> TupleGet(List.length l, translate t, n)
	  | Typeinf.TVar ({ contents = Some typ }) -> translate (Typeinf.TupleGet(typ, t, n))
	  | _ -> failwith "error: translate"
      end
*)
  | Typeinf.Var x -> Var x
  (* | Typeinf.Fix(fs, ts) -> Fix(fs, List.map translate ts) *)
  | Typeinf.Fix(f, (x, u), t) -> (* Fix([f], [Lam((x,u), translate (fst t))]) *)
      Lam((x,u), translate (fst t))
  (* | Typeinf.Lam(x, u, t) -> Lam((x, u), translate t) *)
  | Typeinf.App(t1, t2) -> App(translate (fst t1), translate (fst t2))
  | Typeinf.Let(x, t1, t2) -> Let(x, translate (fst t1), translate (fst t2))
  (* | Typeinf.Leaf(Typeinf.Plus, t) -> Write(translate t) *)
  (* | Typeinf.Leaf(Alpha.UOmega, t) -> Leaf (translate t) *)
  | Typeinf.Node(Alpha.UPlus, l, t_attr, t) ->
      Seq([Write(StartTag(translate (fst l), translate (fst t_attr))); translate (fst t); Write(CloseTag(translate (fst l)))])
  | Typeinf.Node(Alpha.UOmega, l, t_attr, t) ->
      Node(translate (fst l), translate (fst t_attr), translate (fst t))
  | Typeinf.Label l -> Label l
  | Typeinf.Attribute(attribute) -> (* Attr (List.map (fun (l, (t, typ)) -> (l, translate t)) attribute) *)
      Attr attribute
  | Typeinf.AttrGet(t, id) -> AttrGet(translate (fst t), id)
  | Typeinf.Nil Alpha.UPlus -> Unit
  | Typeinf.Nil Alpha.UOmega -> Nil
  | Typeinf.Cons(Alpha.UPlus, hd, tl) ->
      Seq([translate (fst hd); translate (fst tl)])
  | Typeinf.Cons(Alpha.UOmega, hd, tl) ->
      Cons(translate (fst hd), translate (fst tl))
  | Typeinf.Match(((Alpha.TTreelist Alpha.UOmega) | Alpha.TTree Alpha.UOmega), t, pats) ->
      (* MatchList(translate t, translate t1, hd, tl, translate t2) *)
      Match((* translate t *) (Var t), 
      List.map
	(fun (pat, t) -> 
	  let p, _, _ = translate_pattern pat Alpha.UOmega in
	  (p, translate (fst t))) 
	pats)
  | Typeinf.Match(Alpha.TTreelist Alpha.UOne, t, pats) ->
      let match_nil, t_else =
	try
	  let t_nil, t_else = Alpha.pick_case (Alpha.NilPattern) pats in
	  let l = Alpha.new_id "l" in
	    [EndTagPattern(IdentPattern l),
	     Seq [(* Comment "Case nil list"; *) Read; Assert(BinOp(Typeinf.Eq, Pop, Leaf (Alpha.UOne,(Var l)))); 
	     translate (fst (snd t_nil))]], t_else
	with
	  Failure _ -> [], pats
      in
      let matches_else =
	List.map 
	  (fun (pat, t) ->
	    let vars_in_pat = Alpha.pick_vars pat in
	    let p, fv, _ = translate_pattern pat Alpha.UOne in
	    p, (List.fold_right (fun x t' ->substitute x Unit t') vars_in_pat (translate t)))
	  (List.map (fun (p, t) -> (p, fst t)) t_else)
      in
      Match(Seq([Var t; Peek]), matches_else @ match_nil)
  | Typeinf.Match((Alpha.TTree Alpha.UOne), t, pats) ->
      let match_pcdata, t_else =
	try
	  let t_pcdata, t_else = Alpha.pick_case (Alpha.LeafPattern(Alpha.IdentPattern (Alpha.new_id "_"))) pats in
	  let p, _, _ = translate_pattern (fst t_pcdata) Alpha.UOne in
	  begin
	    match fst t_pcdata with
	      Alpha.LeafPattern (Alpha.IdentPattern (x)) ->
		[p, Seq[Read; (* substitute x (Pcdata (Alpha.UOne, (Var x))) *) (translate (fst (snd t_pcdata)))]], t_else
	  end
	with
	    Failure _ -> [], pats
      in
      let match_ident, t_else =
	try
	  let t_ident, t_else = Alpha.pick_case (Alpha.IdentPattern (Alpha.new_id "_")) t_else in
	  let (IdentPattern x), _, _ = translate_pattern (fst t_ident) Alpha.UOne in
	  [(IdentPattern x), substitute x Unit (translate (fst (snd t_ident)))], t_else
	with
	    Failure _ -> [], t_else
      in
      let match_else =
	List.map
	  (fun (pat, t) ->
	    let vars_in_pat = Alpha.pick_vars pat in
	    let p, fv, labs = translate_pattern pat Alpha.UOne in
	     let seq_push = 
	       List.fold_right 
		 (fun l t -> match l with `Lab l -> Seq [Read; (Push (Leaf (Alpha.UOne, (String l)))); t] | `Var x -> Seq [Read; (Push (Leaf (Alpha.UOne, (Var x)))); t]) 
		 labs 
		 Unit 
	     in
	     let t' = List.fold_right (fun x t' -> substitute x Unit t') vars_in_pat (translate t) in
	       p, Seq[seq_push; t'])
	  (* pats *) (List.map (fun (p, t) -> (p, fst t)) t_else)
      in
      Match(Seq[Var t; (* Read *) Peek], match_else @ match_pcdata @ match_ident)
(*
      Match(Seq[translate t; (* Read *) Peek],
	    (List.map
	       (fun (p, t') -> 
		  let p, fv, lab = translate_pattern p in
		  let t' = 
		    List.fold_right 
		      (fun x t' -> substitute x Unit t')
		      fv
		      (translate t')
		  in
		    match lab with
			[] -> p, t' (* begin match p with PcdataPattern p' -> p, t' | _ -> p, Seq ([Read; t']) end *)
(*
		      | [hd] -> p, Seq [Push(Pcdata (String hd)); t']
		      | hd::tl ->
			  p, (Seq ((Push (Pcdata (String hd)))::((List.map (fun _ -> Push(Read)) tl) @ [t']))))
*)
		      | _ -> p, (Seq ((List.map (fun l -> Seq[Read; Push (Pcdata (String l))]) lab) @ [t'])))
	       pats))
		    (* translate_pattern p,  pats) *)
*)	
    
(*
  | Typeinf.Match(Typeinf.TTree Alpha.UOne, t, pats) ->
*)
      
      (* Match(Seq([translate t; Read]),  *)
(*
  | Typeinf.MatchTree(Alpha.UOne, t, x, t1, l, al, tl, t2) ->
      MatchStart(
	Seq[translate t; Read],
	l, al, substitute tl Unit (translate t2),
	x, translate t1)
  | Typeinf.MatchTree(Alpha.UOmega, t, x, t1, l, al, tl, t2) ->
      MatchTree(translate t, x, translate t1, l, al, tl, translate t2)
  | Typeinf.MatchList(Alpha.UOne, t, t1, hd, tl, t2) ->
      let l = Typeinf.new_ident () in
	MatchEnd(
	  Seq [translate t; Peek],
	  l, Seq [Read; Assert(BinOp(Typeinf.Eq, Pop, Var l)); translate t1],
	  substitute hd Unit (substitute tl Unit (translate t2)))
*)
  | Typeinf.If(t, t1, t2) -> If (translate (fst t), translate (fst t1), translate (fst t2))
  | Typeinf.BinOp(b, t1, t2) -> BinOp(b, translate (fst t1), translate (fst t2))
  (* | Typeinf.Label l -> Label l *)
  | Typeinf.False -> False
  | Typeinf.True -> True
  | Typeinf.Cons(Alpha.UOne, _, _) | Typeinf.Nil(Alpha.UOne) | Typeinf.Node(Alpha.UOne, _, _, _) -> failwith ""
  | Typeinf.Skip(_, t) -> Seq [App(Var (Alpha.new_id "Runtime.skip_tree"), Unit); translate (fst t)]
  | Typeinf.Copy _ -> App(Var (Alpha.new_id "Runtime.copy_tree"), Unit)
  | Typeinf.M2S_tl t -> App(Var (Alpha.new_id "Runtime.print_list"), translate (fst t))
  | Typeinf.M2S_t t -> App(Var (Alpha.new_id "Runtime.print_tree"), translate (fst t))
  | Typeinf.LetS2M_tl(x, t) ->
      Let((x, Alpha.UOmega), App(Var (Alpha.new_id "Runtime.parse_list"), Unit), translate (fst t))
  | Typeinf.LetS2M_t(x, t) ->
      Let((x, Alpha.UOmega), App(Var (Alpha.new_id "Runtime.parse_tree"), Unit), translate (fst t))
  | Typeinf.Letrecs (defs, t) ->
      Letrecs ((List.map (fun (f, t) -> (f, translate (fst t))) defs), translate (fst t))
      
(*
let rec print chan formatter = function
    Leaf (u, t) -> fprintf chan "%a" (print chan) t
  | String s -> fprintf chan "%s" s
  | True -> fprintf chan "true"
  | False -> fprintf chan "false"
  | Var x -> fprintf chan "%s" x.Alpha.origin
  | BinOp(b, t1, t2) -> fprintf chan "%a %a %a" (print chan) t1 (Typeinf.print_binop chan) b (print chan) t2
  | Assert t -> fprintf chan "assert(%a)" (print chan) t
  | If(t, t1, t2) -> fprintf chan "if %a then %a else %a" (print chan) t (print chan) t1 (print chan) t2
  (* | Fix(fs, ts) -> fprintf chan "let rec %s = %a in %s" f (print chan) t f *)
  | Lam((x, u), t) -> fprintf chan "fun %s -> %a" x (print chan) t
  | App(t1, t2) -> fprintf chan "%a %a" (print chan) t1 (print chan) t2
  | Write t -> fprintf chan "write %a" (print chan) t
  (* | Leaf t -> fprintf chan "Leaf %a" (print chan) t *)
  | Node(l, t, t1) -> 
      fprintf chan "Node(%a, %a, %a)" (print chan) l (print chan) t (print chan) t1
  | StartTag(l, t) ->
      fprintf chan "StartTag(%a, %a)" (print chan) l (print chan) t
  | CloseTag l ->
      fprintf chan "CloseTag(%a)" (print chan) l
  | Seq [] -> ()
  | Seq [hd] -> print chan formatter hd
  | Seq (hd::tl) ->
      fprintf chan "%a;%a" (print chan) hd (print chan) (Seq tl)
  | Unit -> fprintf chan "()"
  | Nil -> fprintf chan "[]"
  | Label l -> fprintf chan "%a" (Typeinf.print_label chan) l
  | AttrGet(t, x) -> fprintf chan "%a.%s" (print chan) t x
  | Attr attr -> fprintf chan "%a" (print_attr chan) attr
  | Cons(hd, tl) -> fprintf chan "%a::%a" (print chan) hd (print chan) tl
  | Push t -> fprintf chan "push %a" (print chan) t
  | Pop -> fprintf chan "pop ()"
  | Read -> fprintf chan "read lexbuf"
  | Peek -> fprintf chan "peek lexbuf"
  | Match(t, pats) ->
      fprintf chan "match %a with %a" (print chan) t (print_pats chan) pats
(*
  | MatchStart(t, l, attr, t1, x, t2) ->
      fprintf chan "match %a with StartTag(%s, %s) -> %a | Pcdata %s -> %a"
	(print chan) t l attr (print chan) t1 x (print chan) t2
  | MatchEnd(t, l, t1, t2) ->
      fprintf chan "match %a with EndTag(%s) -> %a | _ -> %a" 
	(print chan) t l (print chan) t1 (print chan) t2
  | MatchTree(t, x, t1, l, attr, tree, t2) ->
      fprintf chan "match %a with Leaf %s -> %a | Node(%s, %s, %s) -> %a"
	(print chan) t x (print chan) t1 l attr tree (print chan) t2
  | MatchList(t, t1, hd, tl, t2) ->
      fprintf chan "match %a with [] -> %a | %s::%s -> %a"
	(print chan) t (print chan) t1 hd tl (print chan) t2
*)
and print_attr chan formatter = function
    [] -> ()
  | [s, t] -> fprintf chan "{ %s = %a }" s (print chan) t
  | (s, t)::tl -> fprintf chan "{ %s = %a; %a }" s (print chan) t (print_attr chan) tl
and print_pats chan formatter = function
    [] -> ()
  | [p, t] -> fprintf chan "%a -> %a" (print_pattern chan) p (print chan) t
  | (p, t)::tl -> fprintf chan "%a -> %a | %a" (print_pattern chan) p (print chan) t (print_pats chan) tl
and print_pattern chan formatter = function
    NilPattern -> fprintf chan "[]"
  | ConsPattern(p1, p2) -> fprintf chan "%a::%a" (print_pattern chan) p1 (print_pattern chan) p2
  | TreePattern(p, p1, p2) -> fprintf chan "<%a>[%a@%a]" (print_pattern chan) p (print_pattern chan) p2 (print_pattern chan) p1
  | PcdataPattern (_, p) -> fprintf chan "\"%a\"" (print_pattern chan) p
  | StringPattern s -> fprintf chan "%s" s
  | LabelPattern s -> fprintf chan "%s" s
  | StartTagPattern (p, p_a) -> fprintf chan "<%a%a>" (print_pattern chan) p (print_pattern chan) p_a
  | EndTagPattern p -> fprintf chan "</%a>" (print_pattern chan) p
  | IdentPattern x -> fprintf chan "%s" x
  | AttributePattern a ->
      fprintf chan "%a" (print_attribute chan) a
and print_attribute chan formatter = function
    [] -> fprintf chan ""

  | [k, v] -> fprintf chan " %s=%s" k v
  | (k, v)::tl -> fprintf chan " %s=%s %a" k v (print_attribute chan) tl
      
*)
