type ident = Alpha.ident
type pcdata = Alpha.pcdata
type label = Alpha.label
type attribute = Alpha.attribute

type use = Alpha.use

type constr =
    Ge_use of use * use
  | Eq_use of use * use

type env = (ident * Alpha.typ) list

type binop = Eq | Sub | Cat | ListCat | Semi | Mod

(*
let print_binop chan = function
    Eq -> output_string chan "="
  | Sub -> output_string chan "-"
  | Cat -> output_string chan "^"
  | ListCat -> output_string chan "@"
  | Semi -> output_string chan ";"
  | Mod -> output_string chan "mod"
*)

type pre_t =
    String of string
  | Int of int
  | True | False
  (* | Pcdata of pcdata *)
  | Attribute of attribute
  | Label of label
  | AttrGet of t * ident
  | BinOp of binop * t * t
  | Var of ident
  | Fix of ident * (ident * use) * t
  | App of t * t
  | Let of (ident * use) * t * t
  | Letrecs of (ident * t) list * t
  | If of t * t * t
  | Leaf of Alpha.use * t
  | Node of Alpha.use * t * t * t
  | Nil of Alpha.use
  | Cons of Alpha.use * t * t
  | Match of Alpha.typ * ident * one_match list
  | LetS2M_t of ident * t
  | LetS2M_tl of ident * t
  | M2S_t of t
  | M2S_tl of t
  | Copy of ident
  | Skip of ident * t
  | Coerce of env * env * t
  | CoerceOut of Alpha.typ * Alpha.typ * t

and t = pre_t * Alpha.typ
and one_match = pattern * t
and pattern = Alpha.pattern
(* and attribute = (string * t) list *)

let use_of = function
    Alpha.TString
  | Alpha.TInt
  | Alpha.TBool
  | Alpha.TLabel
  | Alpha.TAttribute
  | Alpha.TFun (_, _) -> Alpha.UOmega
  | Alpha.TTree u -> u
  | Alpha.TTreelist u -> u

let rec print_usage u =
  match u with
    Alpha.UOmega -> print_string "Omega"
  | Alpha.UOne -> print_string "One"
  | Alpha.UPlus -> print_string "Plus"
  | Alpha.UVar i -> print_string "u";print_int i
let rec print_typ typ = 
  match typ with
      Alpha.TTree u -> print_string "TTree^{"; print_usage u; print_string "}" 
    | Alpha.TTreelist u -> print_string "TTreelist^{"; print_usage u; print_string "}" 
    | Alpha.TLabel -> print_string "TLabel"
    | Alpha.TAttribute-> print_string "TAttr"
    | Alpha.TFun(t1, t2)->
	print_string "TFun(";
	print_typ t1;
	print_string ", ";
	print_typ t2;
	print_string ")"	
    | Alpha.TString -> print_string "TStr"
    | Alpha.TInt -> print_string "TInt"
    | Alpha.TLabel -> print_string "TLabel"
    | Alpha.TBool -> print_string "TBool"
    | _ -> print_string "other_typ"
let rec print_id id =
  (* print_int id.Alpha.id *)
  print_string id.Alpha.origin
let rec print_env env =
  print_string "{";
  let rec sub_print_env env = 
    match env with
	[] -> print_string "}"
  | (id, ((Alpha.TTree u) as typ))::env' | (id, ((Alpha.TTreelist u) as typ))::env' ->
      print_string "(";
      print_id id;
      print_string ", ";
      print_typ typ;
      print_string ")";
      sub_print_env env'
  | (id, typ)::env' ->
      print_string "(";
      print_id id;
      print_string ", ";
      print_typ typ;
      print_string ")";
      sub_print_env env'
  in
    sub_print_env env
      
let rec print_constr cs =
  match cs with
      [] -> print_string "[]"
    | Ge_use(u1, u2)::c2 ->
	print_string "(";
	print_usage u1;
	print_string ">";
	print_usage u2;
	print_string "):: ";
	print_constr c2
    | Eq_use(u1, u2)::c2 ->
	print_string "(";
	print_usage u1;
	print_string "=";
	print_usage u2;
	print_string "):: ";
	print_constr c2

let rec rename env =
  match env with
      [] -> []
    | (id, typ):: tl ->
	(let typ' =
	  match typ with
	      Alpha.TTree (Alpha.UVar id) -> 
		Alpha.TTree (Alpha.new_uvar ())
	    | Alpha.TTree (Alpha.UOne) ->
		Alpha.TTree (Alpha.new_uvar ())
	    | Alpha.TTreelist (Alpha.UVar id) ->
		Alpha.TTreelist (Alpha.new_uvar ())
	    | Alpha.TTreelist (Alpha.UOne) ->
		Alpha.TTreelist (Alpha.new_uvar ())
	    | _ -> typ
	in
	  (id,typ')::(rename tl))

let rec well_formed env =
  match env with
      [] -> []
    | [id,typ] -> []
    | (id1,typ1)::( ((id2,typ2)::_) as tl ) ->
	let u1 = use_of typ1 in
	let u2 = use_of typ2 in
	  if u1 = u2 then
	    well_formed tl
	  else
	    (Ge_use(u1, u2))::(well_formed tl)

(* Theta' > Theta *)
let rec above_than_below env_above env_below =
  match env_above, env_below with
      [], [] -> []
    | ((id1, typ1)::env_above'), ((id2, typ2)::env_below') when id1.Alpha.id = id2.Alpha.id ->
	let u1 = use_of typ1 in
	let u2 = use_of typ2 in
	  if u1 = u2 then
	    above_than_below env_above' env_below'
	  else
	    (Ge_use(u1, u2))::(above_than_below env_above' env_below')
    | _ -> failwith "above_than_below: two envs don't match."

(* env 中の、xs 以外の変数は全部omega*)
let rec skip_unused env (xs:ident list) =
  match env with
      [] -> []
    | (id,typ)::tl ->
	if List.exists (fun x -> id.Alpha.id = x.Alpha.id) xs then
	  skip_unused tl xs
	else
	  let u = use_of typ in
	    if u = Alpha.UOmega then
	      (skip_unused tl xs)
	    else
	      (Ge_use(u, Alpha.UOmega))::(skip_unused tl xs)

let rec subtract_env env y =
  match env with
      [] -> []
    | (id, typ)::env' ->
	if id = y then (subtract_env env' y)
	else id::(subtract_env env' y)

let rec subtracts_env env ids =
  match env with
      [] -> []
    | (id, typ)::env' ->
	if (List.mem id ids) then (subtracts_env env' ids)
	else id::(subtracts_env env' ids)
      
let rec unify_typ typ1 typ2 =
  match typ1, typ2 with
      Alpha.TString, Alpha.TString
    | Alpha.TInt, Alpha.TInt
    | Alpha.TBool, Alpha.TBool
    | Alpha.TLabel, Alpha.TLabel
    | Alpha.TAttribute, Alpha.TAttribute -> []
    | (Alpha.TFun(typ1d, typ1cod)), (Alpha.TFun(typ2d, typ2cod)) ->
	(unify_typ typ1d typ2d) @ (unify_typ typ1cod typ2cod)
    | (Alpha.TTree u1), (Alpha.TTree u2) ->
	if u1 = u2 then
	  []
	else
	  [Eq_use(u1, u2)]
    | (Alpha.TTreelist u1), (Alpha.TTreelist u2) ->
	if u1 = u2 then
	  []
	else
	  [Eq_use(u1, u2)]
    | _ -> failwith "unify_typ: CANNOT OCCUR!"

let rec restrict env (ids:ident list) =
  match env with
    [] -> []
  | (id,typ)::tl ->
      if List.exists (fun x -> x.Alpha.id = id.Alpha.id) ids (* List.mem id ids *) then
	(id,typ)::(restrict tl ids)
      else
	(restrict tl ids)

(* 型環境の一番右のID *)
let rec rightmost_id env =
  match env with
      [] -> failwith "rightmost_id: CANNOT OCCUR!"
    | [(id, _)] -> id
    | _::tl -> rightmost_id tl

(* 型環境の、 id が x より左のところだけ取り出す*)
let rec left_than x env =
  match env with
      [] -> failwith "left_than: CANNOT OCCUR!" (* [] *)
    | (id,_)::tl when id.Alpha.id = x.Alpha.id -> []
    | (id,typ)::tl ->
	(id,typ)::(left_than x tl)

let rec mem_permut env env1 env2 =
  match env1 with
      [] -> []
    | _ ->
	let rightmost = rightmost_id env1 in
	let env_L = left_than rightmost env in
	let dom_env2 = List.map fst env2 in
	  List.map
	    (fun (id,typ) -> Ge_use(use_of typ, Alpha.UOmega))
	    (List.filter 
	       (fun (id,typ) -> 
		  List.exists (fun x -> id.Alpha.id = x.Alpha.id) dom_env2) 
	       env_L)

let merge_constr env env1 env2 =
  (* c0 : env1にもenv2にもでてこないのはomega *)
  let c0 = skip_unused env ((List.map fst env1) @ (List.map fst env2)) in
  (* c1 : env2の変数がenv1の変数より左にでてたらomega*)
  let c1 = mem_permut env env1 env2 in
  let dom_env1 = List.map fst env1 in
  let dom_env2 = List.map fst env2 in
  let env' = 
    List.filter 
      (fun (id,_) -> (List.mem id dom_env1) && (List.mem id dom_env2)) 
      env
  in
  (* c2 : env1にもenv2にも共通してでてくるのはomega*)
  let c2 = List.map (fun (id,typ) -> Ge_use(use_of typ, Alpha.UOmega)) env' in
    c0 @ c1 @ c2

let rec partition env id =
  match env with
      [] -> [], []
    | (id',typ')::tl when id.Alpha.id = id'.Alpha.id ->
	[], tl
    | (id',typ')::tl ->
	let l, r = partition tl id in
	  ((id',typ')::l), r

let rec translate (env:env) (term:Alpha.t) (typ:Alpha.typ) =
  let term', typ_of_term = term in
  let t, constr =
    match term' with
	Alpha.Var x ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [x] in
	  let typ' = 
	    snd (List.find (fun (id, _) -> id.Alpha.id = x.Alpha.id) env') in
	  let c3 = unify_typ typ' typ in
	    Coerce(env, env', (Var x, typ)), (c0 @ c1 @ c2 @ c3)
      | Alpha.String s ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let c3 = (unify_typ typ Alpha.TString) in
	    (Coerce(env, env', (String s, typ))), (c0 @ c1 @ c2 @ c3)
      | Alpha.Int n ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let c3 = (unify_typ typ Alpha.TInt) in
	    (Coerce(env, env', (Int n, typ))), (c0 @ c1 @ c2 @ c3)
      | Alpha.True ->
	  True, []
      | Alpha.False ->
	  False, []
      | Alpha.Label l ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let c3 = (unify_typ typ Alpha.TLabel) in
	    (Coerce(env, env', (Label l, typ))), (c0 @ c1 @ c2 @ c3)
      | Alpha.Attribute a ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let c3 = (unify_typ typ Alpha.TAttribute) in
	  (Coerce(env, env', (Attribute a, typ))), (c0 @ c1 @ c2 @ c3)
      | Alpha.AttrGet (a, x) ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let a', c3 = translate env' a Alpha.TAttribute in
	    AttrGet(a', x), (c0@c1@c2@c3)
      | Alpha.BinOp(Alpha.Eq, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let typ1 = snd t1 in
		let typ2 = snd t2 in
		let t1', c4 = translate env1' t1 typ1 in
		let t2', c5 = translate env2' t2 typ2 in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ Alpha.TBool) in
		let c8 = (unify_typ typ1 typ2) in
		  BinOp(Eq, (t1'', typ1), (t2'', typ2)) ,
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7 @ c8)
      | Alpha.BinOp(Alpha.Sub, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let t1', c4 = translate env1' t1 Alpha.TInt in
		let t2', c5 = translate env2' t2 Alpha.TInt in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ Alpha.TInt) in
		  BinOp(Sub, (t1'', typ), (t2'', typ)) ,
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7)
      | Alpha.BinOp(Alpha.Cat, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let t1', c4 = translate env1' t1 (Alpha.TTree Alpha.UOmega) in
		let t2', c5 = translate env2' t2 (Alpha.TTree Alpha.UOmega) in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ (Alpha.TTree Alpha.UOmega)) in
		  BinOp(Cat, (t1'', typ), (t2'', typ)),
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7)
      | Alpha.BinOp(Alpha.ListCat, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let t1', c4 = translate env1' t1 (Alpha.TTreelist Alpha.UOmega) in
		let t2', c5 = translate env2' t2 (Alpha.TTreelist Alpha.UOmega) in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ (Alpha.TTreelist Alpha.UOmega)) in
		  BinOp(ListCat, (t1'', typ), (t2'', typ)),
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7)
      | Alpha.BinOp(Alpha.Semi, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let typ1 = snd t1 in
		let typ2 = snd t2 in
		let t1', c4 = translate env1' t1 typ1 in
		let t2', c5 = translate env2' t2 typ2 in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ typ2) in
		  BinOp(Semi, (t1'', typ1), (t2'', typ2)) ,
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7)
      | Alpha.BinOp(Alpha.Mod, t1, t2) ->
		let env1 = restrict env (Alpha.fv t1) in
		let env2 = restrict env (Alpha.fv t2) in
		let env1' = rename env1 in
		let env2' = rename env2 in
		let c0 = well_formed env1' in
		let c1 = well_formed env2' in
		let c2 = above_than_below env1' env1 in
		let c3 = above_than_below env2' env2 in
		let t1', c4 = translate env1' t1 Alpha.TInt in
		let t2', c5 = translate env2' t2 Alpha.TInt in
		let c6 = merge_constr env env1 env2 in
		let t1'' = Coerce(env1, env1', t1') in
		let t2'' = Coerce(env2, env2', t2') in
		let c7 = (unify_typ typ Alpha.TInt) in
		  BinOp(Mod, (t1'', typ), (t2'', typ)) ,
		(c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7)
      | Alpha.Fix(f, x, body) ->
	    (match typ with
		(Alpha.TFun(typ_d, typ_cod)) as typ_f ->
		  let env' = ((f,typ_f)::env) @ [x,typ_d] in
		  let env'' = rename env' in
(*
		  let c0 = List.map (fun t -> Ge_use(use_of (snd t), Alpha.UOmega), FixEnv) env in
*)
		  let c0 = skip_unused env [] in
		  let c1 = well_formed env'' in
		  let c2 = above_than_below env'' env' in
		  let body', c3 = translate env'' body typ_cod in
		  (Fix(f, (x, use_of typ_d), (Coerce(env', env'', body'), typ_cod))), 
		    (c0 @ c1 @ c2 @ c3)
	      | _ -> failwith "translate: types do not match in function.")
      | Alpha.App(t1, t2) ->
	  let env1 = restrict env (Alpha.fv t1) in
	  let env2 = restrict env (Alpha.fv t2) in
	  let env1' = rename env1 in
	  let env2' = rename env2 in
	  let c0 = well_formed env1' in
	  let c1 = well_formed env2' in
	  let typ_dom = snd t2 in
	  let t1', c2 = translate env1' t1 (Alpha.TFun(typ_dom, typ)) in
	  let t2', c3 = translate env2' t2 typ_dom in
	  let c4 = above_than_below env1' env1 in
	  let c5 = above_than_below env2' env2 in
	  let c6 = merge_constr env env1 env2 in
	  let t1'' = Coerce(env1, env1', t1') in
	  let t2'' = Coerce(env2, env2', t2') in
	    App((t1'', Alpha.TFun(typ_dom, typ)), (t2'', typ_dom)), 
	  (c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6)
      | Alpha.If(t0, t1, t2) ->
	  let env0 = restrict env (Alpha.fv t0) in
	  let env0' = rename env0 in
	  let c0 = well_formed env0' in
	  let c1 = above_than_below env0' env0 in
	  let s =
	    (subtracts_env env (Alpha.fv t0))@
	    (Alpha.fv t1)@
	    (Alpha.fv t2)
	  in
	  let env1 = restrict env s in
	  let env1' = rename env1 in
	  let c2 = well_formed env1' in
	  let c3 = above_than_below env1' env1 in
	  let env2 = restrict env s in
	  let env2' = rename env2 in
	  let c4 = well_formed env2' in
	  let c5 = above_than_below env2' env2 in
	  let t0', c6 = translate env0' t0 (Alpha.TBool) in
	  let t1', c7 = translate env1' t1 typ in
	  let t2', c8 = translate env2' t2 typ in
	  let c9 = merge_constr env env0 env1 in
	  let t1'' = Coerce(env1, env1', t1') in
	  let t2'' = Coerce(env2, env2', t2') in
	  let t = If(t0', (t1'', typ), (t2'', typ)) in
	  let c = c0@c1@c2@c3@c4@c5@c6@c7@c8@c9 in
	    t, c

      | Alpha.Let(x, t1, t2) ->
	  let env1 = restrict env (Alpha.fv t1) in
	  let env1' = rename env1 in
	  let c0 = well_formed env1' in
	  let c1 = above_than_below env1' env1 in
	  let typ_t1 = snd t1 in
	  let t1', c2 = translate env1' t1 typ_t1 in
	  let c7 = [Eq_use(use_of typ_t1, Alpha.UOmega)] in
	  let env2 = (x, typ_t1)::(restrict env (Alpha.fv t2)) in

	  let env2' = rename env2 in
	  let c3 = well_formed env2' in
	  let c4 = above_than_below env2' env2 in
	  let c5 = merge_constr env env1 env2 in
	  let t2', c6  = translate env2' t2 typ in
	  let t2'' = Coerce(env2, env2', t2') in
	  Let((x, use_of typ_t1), t1', (t2'', typ)),
	  (c0@c1@c2@c3@c4@c5@c6@c7)
      | Alpha.Letrecs (fts, t) ->
	  let rec env_letrecs fts env =
	    (match fts with
		 (fn, tn)::tl ->
		   (fn, (snd tn))::(env_letrecs tl env)
	       | [] -> env)
	  in
	  let env' = env_letrecs fts env in
	  let rec trans_letrecs fts =
	    (match fts with
		 (fn, tn)::tl ->
		   let typ_tn = List.assoc fn env' in
		   let tn', c1 = translate env' tn typ_tn in
		     ((fn, tn'), c1)::(trans_letrecs tl)
	       | [] -> [])
	  in
	  let env2 = restrict env' (Alpha.fv t) in
	  let env2' = rename env2 in
	  let c1 = well_formed env2' in
	  let c2 = above_than_below env2' env2 in
	  let t', c3 = translate env' t typ in
	  let fts_cs = trans_letrecs fts in
	  let fts' = List.map fst fts_cs in
	  let cs = List.map snd fts_cs in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	  let t'' = Coerce(env2, env2', t') in
	  let c = c@c1@c2@c3 in
	    Letrecs(fts', (t'', typ)), c    
(*
      | Alpha.Pcdata pc ->
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let c2 = skip_unused env' [] in
	  let c3 = (unify_typ typ Alpha.TString(*TPcdata*)) in
	    (Coerce(env, env', (Pcdata pc, typ))), (c0 @ c1 @ c2 @ c3) *)
      | Alpha.Leaf t ->
	  let use_of_t =
	    match typ with
		Alpha.TTree u -> u
	      | _ -> failwith "translate: types do not match in leaf."
	  in
	  let env' = rename env in
	  let c0 = well_formed env' in
	  let c1 = above_than_below env' env in
	  let t'', c2 = translate env' t Alpha.TString(*Pcdata*) in
	  let t' = Leaf(use_of_t, (Coerce(env, env', t''), Alpha.TString(*TPcdata*))) in
	    t', (Ge_use(use_of_t, Alpha.UOmega))::(c0 @ c1 @ c2)
      | Alpha.Node(t1, t2, t3) ->
	  let use_of_t =
	    match typ with
		Alpha.TTree u -> u
	      | _ -> failwith "translate: types do not match in node."
	  in
	  let env1 = restrict env (Alpha.fv t1) in
	  let env1' = rename env1 in
	  let c0 = well_formed env1' in
	  let c1 = above_than_below env1' env1 in
	  let env2 = restrict env (Alpha.fv t2) in
	  let env2' = rename env2 in
	  let c2 = well_formed env2' in
	  let c3 = above_than_below env2' env2 in
	  let env3 = restrict env (Alpha.fv t3) in
	  let env3' = rename env3 in
	  let c4 = well_formed env3' in
	  let c5 = above_than_below env3' env3 in
	  let t1', c6 = translate env1' t1 Alpha.TLabel in
	  let t2', c7 = translate env2' t2 Alpha.TAttribute in
	  let t3', c8 = translate env3' t3 (Alpha.TTreelist Alpha.UOmega) in
          let envtemp = restrict env ((Alpha.fv t1) @ (Alpha.fv t2)) in
	  let c9 = merge_constr envtemp env1 env2  in
	  let c10 = merge_constr env envtemp env3 in
	  let t1'' = Coerce(env1, env1', t1') in
	  let t2'' = Coerce(env2, env2', t2') in
	  let t3'' = Coerce(env3, env3', t3') in
	  let t = Node(Alpha.UOmega, (t1'', Alpha.TLabel), (t2'', Alpha.TAttribute), (t3'', (Alpha.TTreelist Alpha.UOmega))) in
	  let c = (Ge_use(use_of_t, Alpha.UOmega))::
            	    (c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6 @ c7 @ c8 @ c9 @ c10)
	  in
	    t, c

      | Alpha.Cons(t1, t2) ->
	  let use_of_t =
	    match typ with
		Alpha.TTreelist u -> u
	      | _ -> failwith "translate: types do not match in cons."
	  in
	  let env1 = restrict env (Alpha.fv t1) in
	  let env1' = rename env1 in
	  let c0 = well_formed env1' in
	  let c1 = above_than_below env1' env1 in
	  let env2 = restrict env (Alpha.fv t2) in
	  let env2' = rename env2 in
	  let c2 = well_formed env2' in
	  let c3 = above_than_below env2' env2 in	  
          let t1', c4 = translate env1' t1 (Alpha.TTree Alpha.UOmega) in
	  let t2', c5 = translate env2' t2 (Alpha.TTreelist Alpha.UOmega) in
	  let c6 = merge_constr env env1 env2 in
	  let t = Cons(use_of_t, (Coerce(env1, env1', t1'), (Alpha.TTree Alpha.UOmega)), (Coerce(env2, env2', t2'), (Alpha.TTreelist Alpha.UOmega))) in
	  let c = (Ge_use(use_of_t, Alpha.UOmega))::
	    (c0 @ c1 @ c2 @ c3 @ c4 @ c5 @ c6)
	  in
	    t, c
      | Alpha.Nil ->
	  let use_of_t =
	    match typ with
		Alpha.TTreelist u -> u
	      | _ -> failwith "translate: types do not match in nil."
	  in
	  let t = Nil use_of_t in
	  let c = (Ge_use(use_of_t, Alpha.UOmega))::
                    (skip_unused env [])
	  in 
	    t, c
      | Alpha.Match(typ_match, y, matchs) ->
	  let env1 = restrict env [y] in
	  let env1' = rename env1 in
	  let c0 = well_formed env1' in
	  let s = 
	    let rec fv_match (p, t) =
	      let fv_p = Alpha.fv_pattern p in
		(Alpha.subtracts fv_p (Alpha.fv t))
	    in
	    let fvs = List.map fv_match matchs in
	      match fvs with
		  [] -> []
		| hd::tl -> List.fold_left (@) hd tl
	  in
	  let s = (subtract_env env y)@s in
          let env2l, env2r = partition (restrict env (y::s)) y in
	  let env2 = restrict env s in
	  let c3 = above_than_below env1' env1 in
	  let newtyp = snd (List.find (fun (id,typ) -> id.Alpha.id = y.Alpha.id) env) in
	  let newuse = Alpha.use_of_typ newtyp in
	  let t', c6 = translate env1' (Alpha.Var y, newtyp) newtyp in
	  let c9 = skip_unused env2l [] in
	  let c10 = merge_constr env env1 env2 in
	  let rec translate_match env (p, t) =
	    match p with
	 	Alpha.LeafPattern pc -> 	
		  let env2 =
		    (Alpha.make_pattern_env pc Alpha.TString)@(restrict env s) in
		  let env2' = rename env2 in
		  let c1 = well_formed env2' in
		  let c4 = above_than_below env2' env2 in
		  let t1', c7 = translate env2' t typ in
		  let t1'' = Coerce(env2, env2', t1') in
		    (p, (t1'', typ)), (c1@c4@c7)
	      | Alpha.NodePattern (p1, p2, p3) -> 
		  let env3 = restrict
		    (env2l @ 
		       (y, newtyp)::
		       (Alpha.make_pattern_env p1 Alpha.TLabel)@
		       (Alpha.make_pattern_env p2 Alpha.TAttribute)@
		       (Alpha.make_pattern_env p3 (Alpha.TTreelist newuse))@
		       env2r)
		    (
		      (Alpha.id_of_pattern p1)@
			(Alpha.id_of_pattern p2)@
			(Alpha.id_of_pattern p3)@
			s) in
		  let env3' = rename env3 in
		  let c2 = well_formed env3' in	
		  let c5 = above_than_below env3' env3 in	  
		  let t2', c8 = translate env3' t typ in
		  let t2'' = Coerce(env3, env3', t2') in
		    (p, (t2'', typ)), (c2@c5@c8)
	      | Alpha.ConsPattern(p1, p2) ->
		  let env3 = restrict
		    (env2l @
		       (y, newtyp)::
		       (Alpha.make_pattern_env p1 (Alpha.TTree newuse))@
		       (Alpha.make_pattern_env p2 newtyp)@
		       env2r)
		    (
		      (Alpha.id_of_pattern p1)@
			(Alpha.id_of_pattern p2)@s
		    )
		  in
		  let env3' = rename env3 in
		  let c2 = well_formed env3' in
		  let c5 = above_than_below env3' env3 in
		  let t2', c8 = translate env3' t typ in
		  let t2'' = Coerce(env3, env3', t2') in
		    (p, (t2'', typ)), (c2@c5@c8)
	      | Alpha.NilPattern
	      | Alpha.LabelPattern _ | Alpha.StringPattern _ ->
		  let env2 = (restrict env s) in
		  let env2' = rename env2 in
		  let c1 = well_formed env2' in
		  let c4 = above_than_below env2' env2 in
		  let t1', c7 = translate env2' t typ in
		  let t1'' = Coerce(env2, env2', t1') in
		    (p, (t1'', typ)), (c1@c4@c7)
	      | Alpha.IdentPattern id ->
		  let env3 = restrict
		    (env2l @
		       [y, newtyp]@
		       [id, newtyp]@
		       env2r)
		    (id::s)
		  in
		  let env3' = rename env3 in
		  let c2 = well_formed env3' in
		  let c5 = above_than_below env3' env3 in
		  let t2', c8 = translate env3' t typ in
		  let t2'' = Coerce(env3, env3', t2') in
		    (p, (t2'', typ)), (c2@c5@c8)
	  in 
	  let a = List.map (translate_match env) matchs in
	  let p_ts = List.map fst a in
	  let cs = List.map snd a in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	  let c = c@c0@c3@c6@c9@c10 in
	    Match(newtyp, y, p_ts), c
	      
  in
    (t, typ), constr



let print_binop chan binop =
  match binop with
    Eq -> Format.fprintf chan "="
  | Sub -> Format.fprintf chan "-"
  | Cat -> Format.fprintf chan "^"
  | ListCat -> Format.fprintf chan "@"
  | Semi -> Format.fprintf chan ";"
  | Mod -> Format.fprintf chan " mod "

(*	
let rec print_tab n =
  if (n=0) then ()
  else (print_string "  "; print_tab (n-1))

let rec print_term tab_depth (t, typ_t) =
  let use_t = use_of typ_t in
    print_tab tab_depth;
    match t with
	String s ->
	  Printf.printf "\"%s\"" s
      | Int n ->
	  Printf.printf "%d" n
(*
      |	Pcdata pc ->
	  Printf.printf "\"%s\"" pc
*)
      | Label l ->
	  Printf.printf "<%s>" l
      | Attribute a ->
	  let rec print_attr attr =
	    (match attr with
		 [] -> ()
	       | [str, t] ->
		   (Printf.printf "%s=%s" str (Syntax.term_to_string t))
	       | (str, t)::tl ->
		   (Printf.printf "%s=%s; " str (Syntax.term_to_string t); print_attr tl))
	     in
	      Printf.printf "{";
	      print_attr a;
	      Printf.printf "}"

      | AttrGet (a, x) ->
	  print_term 0 a;
	  Printf.printf ".";
	  Printf.printf "%s" x.Alpha.origin
      | True ->
	  Printf.printf "true"
      | False ->
	  Printf.printf "false"
      | Var x ->
	  Printf.printf "%s" (x.Alpha.origin)
      | BinOp(binop, t1, t2) ->
	  print_term 0 t1;
	  print_binop binop;
	  print_term 0 t2

      | App((Fix(f, (x, u), t2), (Alpha.TFun(typ1, typ2))), t1) ->
	  (if (f.Alpha.origin) = "" then
	    print_term 0 (Let((x,u), t1, t2), typ2)
	   else
	    (print_term tab_depth (Fix(f, (x, u), t2), (Alpha.TFun(typ1, typ2)));
	      print_char ' ';
	      print_term 0 t1))
      | App(t1, t2) ->
	  Printf.printf "(";
	  print_term 0 t1;
	  print_char ' ';
	  print_term 0 t2;
	  Printf.printf ")"
 
      | Fix(f, (x, _), t) ->
	  if (f.Alpha.origin) = "" then
	    (Printf.printf "(fun %s -> \n(" x.Alpha.origin;
	     print_term 0 t;
	     Printf.printf "))")
	  else
	    (	  
	   Printf.printf "fix(%s, %s, (" f.Alpha.origin x.Alpha.origin;
	   print_term 0 t;
	   Printf.printf "))")
(*
      | Fix(f, x, t) ->
	  if (f.Alpha.origin) = "" then
	    (printf "fun %s -> \n(" x.Alpha.origin;
	     print_term 0 t;
	     Print.printf ")")
	  else
	    (printf "let rec %s = fun %s -> \n(" f.Alpha.origin x.Alpha.origin;
	     print_term 0 t;
	     Print.printf "))")
*)
      | Let((x,_), t1, t2) ->
	  Printf.printf "let %s = " (x.Alpha.origin);
	  print_term 0 t1;
	  Printf.printf "\n";
	  print_tab tab_depth;
	  Printf.printf "in\n";
	  print_term tab_depth t2
      | Letrecs(fts, t) ->
	  let rec print_letrec fts tab_depth=
	    (match fts with
		 [f, t] -> 
		   Printf.printf "%s = \n" f.Alpha.origin;
		   print_term tab_depth t
	       | (f, t)::tl ->
		   Printf.printf "%s = " f.Alpha.origin;
		   print_term 0 t;
		   Printf.printf "\n";
		   print_tab tab_depth;
		   Printf.printf "and ";
		   print_letrec tl tab_depth) 
	  in
	    Printf.printf "let rec ";
	    print_letrec fts tab_depth;
	    Printf.printf "\n";
	    print_tab tab_depth;
	    Printf.printf "in\n";
	    print_term tab_depth t
(* *)
      | If(t0, t1, t2) ->
	  Printf.printf "if "; 
	  print_term 0 t0;
	  Printf.printf " then\n";
	  print_term (tab_depth+1) t1;
	  Printf.printf "\n";
	  print_tab tab_depth;
	  Printf.printf "else\n";
	  print_term (tab_depth+1) t2
      | Leaf(u, t) ->
	  begin
	    match u with
		Alpha.UPlus -> Printf.printf "(PC "; print_term 0 t; Printf.printf ")"
	      | Alpha.UOmega -> Printf.printf "(~PC "; print_term 0 t; Printf.printf "~)"
	      (* for debug *)
	      | Alpha.UVar use -> Printf.printf "uleaf^"; Alpha.print_usage u; Printf.printf"("; print_term 0 t; Printf.printf ")"
	      | Alpha.UOne -> failwith "PC n ^ UOne Cannot Occur"
	  end
      | Node(u, t1, t2, t3) ->
	  begin
	    match u with
	      | Alpha.UPlus ->
		  Printf.printf "(";
		  print_term 0 t1;
		  Printf.printf "@";
		  print_term 0 t2;
		  Printf.printf "$";
		  print_term 0 t3;
		  Printf.printf ")"
	      | Alpha.UOmega ->
		  Printf.printf "(~";
		  print_term 0 t1;
		  Printf.printf "@";
		  print_term 0 t2;
		  Printf.printf "$";
		  print_term 0 t3;
		  Printf.printf "~)"
	      (* for debug *)
	      | Alpha.UVar use ->
		  Printf.printf "unode^";
		  Alpha.print_usage u;
		  Printf.printf "(";
		  print_term 0 t1; 
		  Printf.printf ", ";
		  print_term 0 t2;
		  Printf.printf ", ";
		  print_term 0 t3;
		  print_char ')'
	      | Alpha.UOne ->
		  print_term 0 t1;
		  Printf.printf "@";
		  print_term 0 t2;
		  Printf.printf "$";
		  print_term 0 t3    

	  end
      | Cons(u, t1, t2) ->
	  begin
	    match u with
	      | Alpha.UPlus ->
		  Printf.printf "(";
		  print_term 0 t1;
		  Printf.printf "::";
		  print_term 0 t2;
		  Printf.printf ")"
	      | Alpha.UOmega ->
		  Printf.printf "(";
		  print_term 0 t1;
		  Printf.printf "%%%%";
		  print_term 0 t2;
		  Printf.printf ")"
	      (* for debug *)
	      | Alpha.UVar use ->
		  Printf.printf "ucons^";
		  Alpha.print_usage u;
		  Printf.printf "(";
		  print_term 0 t1;
		  Printf.printf ", ";
		  print_term 0 t2;
		  print_char ')'		   
	      | Alpha.UOne ->
		  Printf.printf "(";
		  print_term 0 t1;
		  Printf.printf "::";
		  print_term 0 t2;
		  Printf.printf ")"
	  end
      | Nil(u) ->
	  begin
	    match u with
		Alpha.UPlus ->
		  Printf.printf "[]"
	      | Alpha.UOmega ->
		  Printf.printf "<>"
	      | Alpha.UVar use ->
		  Printf.printf "unil^";
		  Alpha.print_usage u;
	      | Alpha.UOne ->
		  Printf.printf "[]"
	  end
      | Match(typ, y, matchs) ->
	  let buf =
	    match (Alpha.use_of_typ typ) with
		Alpha.UOmega -> true
	      | _ -> false
	  in
	  let rec print_pattern p =
	    (
	      match p with
		  Alpha.LeafPattern s ->
		    if buf then
		      (	Printf.printf "~PC ";
			print_pattern s;
			Printf.printf "~")
		    else
		      ( Printf.printf "PC ";
			print_pattern s;
			Printf.printf "")
		| Alpha.NodePattern (p1, p2, p3) ->
		    if buf then
		      (Printf.printf "~";
		       print_pattern p1;
		       Printf.printf "@";
		       print_pattern p2;
		       Printf.printf "$";
		       print_pattern p3;
		       Printf.printf "~")
		    else
		      (print_pattern p1;
		       Printf.printf "@";
		       print_pattern p2;
		       Printf.printf "$";
		       print_pattern p3)
		| Alpha.NilPattern ->
		    if buf then
		      Printf.printf "<>"
		    else
		      Printf.printf "[]"
		| Alpha.ConsPattern (p1,p2)-> 
		    if buf then
		      (print_pattern p1;
		       Printf.printf "%%%%";
		       print_pattern p2)
		    else 
		      (print_pattern p1;
		       Printf.printf "::";
		       print_pattern p2)
		| Alpha.StringPattern s ->
		    Printf.printf "%s" s
		| Alpha.LabelPattern l ->
		    Printf.printf "<%s>" l
		| Alpha.IdentPattern id -> 
		    Printf.printf "%s" id.Alpha.origin
	    )
	  in
	  let rec print_one_match tab_depth (p, t) =
	    print_pattern p;
	    Printf.printf " -> ";
	    print_term 0 t;
	  in
	  let rec print_matchs tab_depth ms =
	    (
	      match ms with
		  [] -> ()
		| hd::tl ->
		    Printf.printf "\n";
		    print_tab tab_depth;
		    Printf.printf "| ";
		    print_one_match tab_depth hd;
		    print_matchs tab_depth tl
	    )
	  in
  	    Printf.printf "match %s with\n" (y.Alpha.origin);
	    let ((p, t)::matchs') = matchs in
	      print_tab (tab_depth+1);
	      Printf.printf "  ";
	      print_pattern p;
	      Printf.printf " -> ";
	      print_term 0 t;
	      print_matchs (tab_depth+1) matchs'
      | LetS2M_t(id, t) -> 
	  if (id.Alpha.origin="_") then print_term 0 t
	  else
	    (Printf.printf "(let %s = s2m_t (%s) in " id.Alpha.origin id.Alpha.origin; print_term 0 t; Printf.printf ")")
      | LetS2M_tl(id, t) -> 
	  if (id.Alpha.origin="_") then print_term 0 t
	  else
	    (Printf.printf "(let %s = s2m_tl (%s) in " id.Alpha.origin id.Alpha.origin; print_term 0 t; Printf.printf ")")
      | M2S_t(t) ->
	  Printf.printf "m2s_t("; print_term 0 t; Printf.printf ")"
      | M2S_tl(t) ->
	  Printf.printf "m2s_tl("; print_term 0 t; Printf.printf ")"
      | Copy(x) ->
	  if (x.Alpha.origin="_") then () else Printf.printf "(copy_tree %s)" x.Alpha.origin
      | Skip(x, t) ->
	  if (x.Alpha.origin="_") then print_term 0 t else
	    (Printf.printf "(skip_tree %s);" x.Alpha.origin; print_term 0 t)
      | Coerce(env_s, env_d, t) -> Printf.printf "[coerce] "; Printf.printf "("; print_env env_s; Printf.printf ") => ("; print_env env_d; Printf.printf ") "; print_term 0 t
      | CoerceOut(typ1, typ2, t) ->
	  Printf.printf "coerceout {";
	  print_typ typ1;
	  Printf.printf ", ";
	  print_typ typ2;
	  Printf.printf "}";
	  print_term 0 t
*)
