type typ =
    TString
  | TInt
  | TBool
  (* | TPcdata *)
  | TLabel
  | TAttribute
  | TFun of typ * typ
  | TTree
  | TTreelist
  | TTuple of typ list
  | TTupleGet of typ * int
  | TVar of typ option ref

type ident = string
type label = string
type pcdata = string

type binop = Eq | Cat | Sub | ListCat | Semi | Mod

type pre_t =
    String of string
  | Int of int
  | True | False
  (* | Pcdata of pcdata *)
  | Attribute of attribute
  | Label of label
  | BinOp of binop * t * t
  | Var of ident
  | Fix of ident * ident * t
  | App of t * t
  | Leaf of t
  | Node of t * t * t
  | Nil
  | Cons of t * t
  | Match of t * one_match list
  | If of t * t * t
  | Let of ident * t * t
  | Letrecs of (ident * t) list * t
  | AttrGet of t * ident
and t = pre_t * typ
and attribute = (string * Syntax.t) list
and one_match = (pattern * t)
and pattern =
    LeafPattern of (*pcdata*) pattern
  | NodePattern of (*label*) pattern * (*attribute*)pattern * (*treelist*)pattern
  | NilPattern
  | ConsPattern of (*tree*)pattern * (*treelist*)pattern
  | IdentPattern of ident
  | StringPattern of string
  | LabelPattern of label

let rec print_pattern p =
  (
  match p with
      LeafPattern _ ->
	print_string "leaf"
    | NodePattern _ ->
	print_string "node"
    | NilPattern ->
	print_string "nil"
    | ConsPattern _ ->
	print_string "cons"
    | StringPattern _ ->
	print_string "str"
    | IdentPattern id ->
	print_string "id_"; print_string id;
    | LabelPattern _ ->
	print_string "label"
  );
  print_newline()

let rec print_typ typ =
  match typ with
    TString -> print_string "TString"
  | TInt-> print_string "TInt"
  | TBool-> print_string "TBool"
   (* | TPcdata -> print_string "TPcdata"*)
  | TLabel-> print_string "TLabel"
  | TAttribute-> print_string "TAttr"
  | TFun (typ1, typ2) ->
      print_string "TFun(";
      print_typ typ1;
      print_string ", ";
      print_typ typ2;
      print_string ")"
  | TTree -> print_string "TTree"
  | TTreelist -> print_string "TTreelist"
  | TVar ({ contents = None} ) -> print_string "TVar_N"
  | TVar ({ contents = Some typ }) ->
      print_string "TVar ";
      print_typ typ
  | _ -> print_string "tuple tukauna"


let rec print_env env =
  match env with
      [] -> print_newline()
    | (id, typ)::tl ->
	print_string "(";
	print_string id;
	print_string ", ";
	print_typ typ;
	print_string "), ";
	print_env tl

let rec print_t (t, typ) =
  print_string "(";
  begin
    match t with
	Var id -> print_string id
      | Fix(f, x, t) when f = "" ->
	  print_string "fun ";
	  print_string x;
	  print_string " -> ";
	  print_t t
      | True ->
	  print_string "true"
      | False ->
	  print_string "false"
      | App(t1, t2) ->
	  print_t t1;
	  print_string " ";
	  print_t t2
      | If(t0, t1, t2) ->
	  print_string "if ";
	  print_t t0;
	  print_string " then ";
	  print_t t1;
	  print_string " else ";
	  print_t t2
      | Let(x, t1, t2) ->
	  print_string "let ";
	  print_string x;
	  print_string " = ";
	  print_t t1;
	  print_string " in ";
	  print_t t2
      | String s -> print_string s
      | BinOp(Eq, t1, t2) ->
	  print_t t1;
	  print_string " = ";
	  print_t t2
      | Nil -> 
	  print_string "[]"
      | Letrecs(fts, t) ->
	  let print_ft (f, t) =
	    print_string "let rec ";
	    print_string f;
	    print_string " = ";
	    print_t t
	  in
	  let rec print_fts fts =
	    begin
	      match fts with
		| [ft] -> print_ft ft
		| ft::tl ->
		    print_ft ft; print_string " and ";  print_fts tl
	    end
	  in
	    print_fts fts;
	    print_string "\nin";
	    print_t t
      | _ -> print_string "other"
  end;
  print_string ", ";
  print_typ typ;
  print_string ")"
  




exception Unify of typ * typ

let new_tvar () = TVar (ref None)

let rec shorten t =
  match t with
      TVar({ contents = Some (TVar { contents = Some t })} as r1) ->
	let t' = shorten t in
	  r1 := Some t'; t'
    | TFun(tdom, tcod) -> TFun(shorten tdom, shorten tcod)
    | _ -> t

let rec unify typ1 typ2 =
  match shorten typ1, shorten typ2 with
       TString, TString | TInt, TInt | TBool, TBool
    | TLabel, TLabel | TAttribute, TAttribute 
    | TTree, TTree | TTreelist, TTreelist -> ()
    (*| Tuple, TupleGet *)
    | TFun(typ1dom, typ1cod), TFun(typ2dom, typ2cod) ->
	unify typ1dom typ2dom;
	unify typ1cod typ2cod
    | TVar r1, TVar r2 when r1 == r2 -> ()
    | TVar { contents = Some(typ1') }, _ -> unify typ1' typ2
    | _, TVar { contents = Some(typ2') } -> unify typ1 typ2'
    | TVar ({ contents = None } as r1), _ -> r1 := Some typ2
    | _, TVar ({ contents = None } as r2) -> r2 := Some typ1
    | _ -> raise (Unify (typ1, typ2))


let rec infer env t =
(*
print_string "env:\n";
print_env env;
print_newline();
print_string (Syntax.term_to_string t);
print_newline();
*)
  match t with 
      Syntax.String s -> (String s), TString
    | Syntax.Int n -> (Int n), TInt
    | Syntax.True -> True, TBool
    | Syntax.False -> False, TBool
    (* | Syntax.Pcdata pc -> (Pcdata pc), TString *)
    | Syntax.Attribute a -> (Attribute a), TAttribute
    | Syntax.Label l -> (Label l), TLabel
    | Syntax.AttrGet(a, x) ->
	let a', typ = infer env a in
	  unify typ TAttribute;
	  AttrGet((a', typ), x), TString
    | Syntax.BinOp(Syntax.Eq, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2' ,typ2 = infer env t2 in
	let typvar = new_tvar() in
	  unify typ1 typvar;
	  unify typ2 typvar;
	  BinOp(Eq, (t1', typ1), (t2', typ2)), TBool
    | Syntax.BinOp(Syntax.Sub, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer env t2 in
	  unify typ1 TInt;
	  unify typ2 TInt;
	  BinOp(Sub, (t1', typ1), (t2', typ2)), TInt
    | Syntax.BinOp(Syntax.Cat, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer env t2 in
	  unify typ1 TTree;
	  unify typ2 TTree;
	  BinOp(Cat, (t1', typ1), (t2', typ2)), TTree
    | Syntax.BinOp(Syntax.ListCat, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer env t2 in
	  unify typ1 TTreelist;
	  unify typ2 TTreelist;
	  BinOp(ListCat, (t1', typ1), (t2', typ2)), TTreelist
    | Syntax.BinOp(Syntax.Semi, t1, t2) -> 
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer env t2 in
	  BinOp(Semi, (t1', typ1), (t2', typ2)), typ2
    | Syntax.BinOp(Syntax.Mod, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer env t2 in
	  unify typ1 TInt;
	  unify typ2 TInt;
	  BinOp(Mod, (t1', typ1), (t2', typ2)), TInt
    | Syntax.Var x -> 
	(Var x), (List.assoc x env)
    | Syntax.Fix(f, x, body) ->
	let typ_dom = new_tvar () in
	let typ_codom = new_tvar () in
	let typ_f = TFun(typ_dom, typ_codom) in
	let typ_x = typ_dom in
	let body', typ_body = infer ((f, typ_f)::(x, typ_x)::env) body in
	  unify typ_body typ_codom;
          Fix(f, x, (body', typ_body)), typ_f
    | Syntax.Lam(x, t) ->
	infer env (Syntax.Fix("", x, t))
    | Syntax.App(t1, t2) ->
	let dom = new_tvar () in
	let codom = new_tvar () in
	let t1', typ_t1 = infer env t1 in
	let t2', typ_t2 = infer env t2 in
	  unify typ_t1 (TFun(dom,codom));
	  unify typ_t2 dom;
	  App((t1', typ_t1), (t2', typ_t2)), codom
(*
    | Syntax.Let(x, t1, t2) ->
	(* infer env (Syntax.App(Syntax.Lam(x, t2), t1)) *)
	infer env (Syntax.App(Syntax.Fix("", x, t2), t1))
*)
(*
    | Syntax.Letrecs(fts, t) ->
	(match fts with
	    [f0, Syntax.Lam(x, t0)] -> 
	      infer env (Syntax.App(Syntax.Lam(f0, t), Syntax.Fix(f0,x,t0)))
	  | _ -> failwith "a")
*)
    | Syntax.Let(x, t1, t2) ->
	let t1', typ_t1 = infer env t1 in
	let t2', typ_t2 = infer ((x, typ_t1)::env) t2 in
	  Let(x, (t1', typ_t1), (t2', typ_t2)), typ_t2
    | Syntax.Letrecs(fts, t) ->
	let rec letrecs_env fts env =
	  (match fts with
	       (fn, tn)::tl ->
		 let typ_fn = new_tvar () in
		   (fn, typ_fn)::(letrecs_env tl env)
	     | [] -> env)
	in
	let env' = letrecs_env fts env in
	let rec infer_letrec fts = 
	  (match fts with
	       (fn, tn)::tl -> 
		 (fn, (infer env' tn))::(infer_letrec tl)
	     | [] -> [])
	in
	let fts' = infer_letrec fts in
	let rec make_env fts env =
	  match fts with
	      (f, (term, typ))::tl ->
		(f, typ)::(make_env tl env)
	    | [] -> env
	in
	let env' = make_env fts' env in
	let (term_t', typ_t') = infer env' t in
	  Letrecs(fts', (term_t', typ_t')), typ_t'
(*
    | Syntax.Letrecs(fts, t) ->
	let rec letrecs_env fts env =
	  (match fts with
	      (f, Syntax.Lam(x, t0))::tl -> 
		let typ_dom = new_tvar() in
		let typ_codom = new_tvar () in
		let typ_f = TFun(typ_dom, typ_codom) in
		let typ_x = typ_dom in
		  (f, typ_f)::(x, typ_x)::(letrecs_env tl env)
	    | [] -> env)
	in
	let env' = letrecs_env fts env in 
	let rec infer_letrec fts env =
	  (match fts with
	       (f0, Syntax.Lam(x, t0))::tl ->
		 (f0, (infer env (Syntax.Lam(x, t0))))::(infer_letrec tl env)
	     | [] -> [])
	in
	let fts' = infer_letrec fts env' in
	let (term_t', typ_t') = infer env' t in
	  Letrecs(fts', (term_t', typ_t')), typ_t' 
*)
(*		 infer env (Syntax.App(Syntax.Lam(f0, t), Syntax.Fix(f0,x,t0)))*)
(*    | Syntax.Let(x, t1, t2) ->
	let t1', typ1 = infer env t1 in
	let t2', typ2 = infer ((x, typ1)::env) t2 in
	  Let (x, (t1', typ1), (t2', typ2)), typ2
*)
    | Syntax.Leaf t ->
	let t', typ_t = infer env t in
	  unify TString typ_t;
	  (Leaf (t', typ_t)), TTree
    | Syntax.Node(t1, t2, t3) -> 
	let t1', typ_t1 = infer env t1 in
	let t2', typ_t2 = infer env t2 in
	let t3', typ_t3 = infer env t3 in
	  unify TLabel typ_t1;
	  unify TAttribute typ_t2;
	  unify TTreelist typ_t3;
	  (Node((t1', typ_t1), (t2', typ_t2), (t3', typ_t3)), TTree)
    | Syntax.Cons(t1, t2) ->
      let t1', typ_t1 = infer env t1 in
      let t2', typ_t2 = infer env t2 in
        unify TTree typ_t1;
        unify TTreelist typ_t2;
        (Cons((t1', typ_t1), (t2', typ_t2)), TTreelist)
    | Syntax.Nil -> (Nil, TTreelist)
    | Syntax.Match(t_match, matchs) ->
	let matchs' = List.map conv_pattern matchs in
	let t_match', typ_match = infer env t_match in
	let temp = List.map (infer_match typ_match env) matchs' in
	let typs_ps = List.map fst temp in
	let p_ts = List.map snd temp in 
	let ts = List.map snd p_ts in
	let typs_ts = List.map snd ts in
	let rec unify_list typ_list = 
	  match typ_list with 
	      [] -> () 
	    | [typ1] -> ()
	    | typ1::typ2::tl -> 
		unify typ1 typ2; 
		unify_list (typ2::tl) 
	in
	  unify_list (typ_match::typs_ps);
	  unify_list typs_ts;
	  let typ_p = List.hd typs_ps in
	  let typ_t = List.hd typs_ts in
	    (Match((t_match', typ_p), p_ts), typ_t)
    | Syntax.If (t0, t1, t2) ->
        let t0', typ0 = infer env t0 in
        let t1', typ1 = infer env t1 in
        let t2', typ2 = infer env t2 in
        let typvar = new_tvar() in
          unify typ0 TBool;
          unify typ1 typvar;
          unify typ2 typvar;
          (If((t0', typ0), (t1', typvar), (t2', typvar)), typvar)
    | _ -> failwith "Not match in typing.infer"
and conv_pattern (p, t) =
  let rec subfun p =
    match p with
	  Syntax.LeafPattern pc -> LeafPattern (subfun pc)
	| Syntax.NodePattern(p1, p2, p3) -> NodePattern(subfun p1, subfun p2, subfun p3)
	| Syntax.NilPattern -> NilPattern
	| Syntax.ConsPattern(p1, p2) -> ConsPattern(subfun p1, subfun p2)
	| Syntax.StringPattern s -> StringPattern s
	| Syntax.IdentPattern id -> IdentPattern id
	| Syntax.LabelPattern l -> LabelPattern l
  in 
  (subfun p, t)
and infer_match typ_match env (pattern, t) =
  let rec make_pattern_env p typ =
    match p with
	LeafPattern pc -> make_pattern_env pc TString
      | NodePattern (p1, p2, p3) ->
	  (make_pattern_env p1 TLabel)@(make_pattern_env p2 TAttribute)@(make_pattern_env p3 TTreelist)
      | NilPattern -> []
      | ConsPattern (p1, p2) ->
	  (make_pattern_env p1 TTree)@(make_pattern_env p2 TTreelist)
      | StringPattern s -> []
      |	IdentPattern id -> [(id, typ)]
      | LabelPattern l -> []
  in
    match pattern with
	LeafPattern pc -> 
	  let env' = (make_pattern_env pattern TTree)@env in
	  let t', typ_t = infer env' t in
	    TTree, (pattern, (t', typ_t))
      | NodePattern (p_label, p_attr, p_treelist) ->
	  let env' = (make_pattern_env  pattern TTree)@env in
	  let t', typ_t = 
	    infer env' t in
	    TTree, (pattern, (t', typ_t))
      | NilPattern -> 
	  let t', typ_t = infer env t in
	TTreelist, (pattern, (t', typ_t))
      | ConsPattern(p1, p2) -> 
	  let env' =
	    (make_pattern_env p1 TTree)@(make_pattern_env p2 TTreelist)@env
	  in
	  let t', typ_t =
	    infer env' t in
	    TTreelist, (pattern, (t', typ_t))
      | StringPattern p -> 
	  let t', typ_t = infer env t in
	    TString, (pattern, (t', typ_t))
      | IdentPattern id ->
	  let t', typ_t = infer ((id, typ_match)::env) t in
	typ_match, (pattern, (t', typ_t))
      | LabelPattern l -> 
	  let t', typ_t = infer env t in
	    TLabel, (pattern, (t', typ_t))
